"""Set based matchers module."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals
from future.utils import python_2_unicode_compatible

from splitio.models.grammar.matchers.base import Matcher


class ContainsAllOfSetMatcher(Matcher):
    """Matcher that returns true if the user data is a subset of the matcher's data."""

    def _build(self, raw_matcher):
        """
        Build an ContainsAllOfSetMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        try:
            setkey = set(matching_data)
            return self._whitelist.issubset(setkey)
        except TypeError:
            return False

    def _add_matcher_specific_properties_to_json(self):
        """Return ContainsAllOfSet specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(list(self._whitelist))
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'contains all of the following set: [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class ContainsAnyOfSetMatcher(Matcher):
    """Matcher that returns true if the intersection of both sets is not empty."""

    def _build(self, raw_matcher):
        """
        Build an ContainsAnyOfSetMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        try:
            return len(self._whitelist.intersection(set(matching_data))) != 0
        except TypeError:
            return False

    def _add_matcher_specific_properties_to_json(self):
        """Return ContainsAnyOfSet specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'contains on of the following se: [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class EqualToSetMatcher(Matcher):
    """Matcher that returns true if the set provided by the user is equal to the matcher's one."""

    def _build(self, raw_matcher):
        """
        Build an EqualToSetMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        try:
            return self._whitelist == set(matching_data)
        except TypeError:
            return False

    def _add_matcher_specific_properties_to_json(self):
        """Return EqualToSet specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'equals the following set: [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class PartOfSetMatcher(Matcher):
    """a."""

    def _build(self, raw_matcher):
        """
        Build an PartOfSetMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        try:
            setkey = set(matching_data)
            return len(setkey) > 0 and setkey.issubset(set(self._whitelist))
        except TypeError:
            return False

    def _add_matcher_specific_properties_to_json(self):
        """Return PartOfSet specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'is a subset of the following set: [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )



"""Miscelaneous matchers that don't fall into other categories."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import json
from future.utils import python_2_unicode_compatible
from six import string_types

from splitio.models.grammar.matchers.base import Matcher


class DependencyMatcher(Matcher):
    """Matcher that returns true if the user's key secondary evaluation result matches."""

    def _build(self, raw_matcher):
        """
        Build an DependencyMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._split_name = raw_matcher['dependencyMatcherData']['split']
        self._treatments = raw_matcher['dependencyMatcherData']['treatments']

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        evaluator = context.get('evaluator')
        assert evaluator is not None

        bucketing_key = context.get('bucketing_key')

        result = evaluator.evaluate_feature(self._split_name, key, bucketing_key, attributes)
        return result['treatment'] in self._treatments

    def _add_matcher_specific_properties_to_json(self):
        """Return Dependency specific properties."""
        return {
            'dependencyMatcherData': {
                'split': self._split_name,
                'treatments': self._treatments
            }
        }


class BooleanMatcher(Matcher):
    """Matcher that returns true if the user submited value is similar to the stored boolean."""

    def _build(self, raw_matcher):
        """
        Build an BooleanMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._data = raw_matcher['booleanMatcherData']

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        if isinstance(matching_data, bool):
            decoded = matching_data
        elif isinstance(matching_data, string_types):
            try:
                decoded = json.loads(matching_data.lower())
                if not isinstance(decoded, bool):
                    return False
            except ValueError:
                return False
        else:
            return False

        return decoded == self._data

    def _add_matcher_specific_properties_to_json(self):
        """Return Boolean specific properties."""
        return {'booleanMatcherData': self._data}



"""Manager tests."""

import pytest
import threading

from splitio.tasks.split_sync import SplitSynchronizationTask
from splitio.tasks.segment_sync import SegmentSynchronizationTask
from splitio.tasks.impressions_sync import ImpressionsSyncTask, ImpressionsCountSyncTask
from splitio.tasks.events_sync import EventsSyncTask
from splitio.tasks.telemetry_sync import TelemetrySynchronizationTask

from splitio.sync.split import SplitSynchronizer
from splitio.sync.segment import SegmentSynchronizer
from splitio.sync.impression import ImpressionSynchronizer, ImpressionsCountSynchronizer
from splitio.sync.event import EventSynchronizer
from splitio.sync.telemetry import TelemetrySynchronizer
from splitio.sync.synchronizer import Synchronizer, SplitTasks, SplitSynchronizers
from splitio.sync.manager import Manager

from splitio.storage import SplitStorage
from splitio.api import APIException


class ManagerTests(object):
    """Synchronizer Manager tests."""

    def test_error(self, mocker):
        split_task = mocker.Mock(spec=SplitSynchronizationTask)
        split_tasks = SplitTasks(split_task, mocker.Mock(), mocker.Mock(), mocker.Mock(),
                                 mocker.Mock(), mocker.Mock())

        storage = mocker.Mock(spec=SplitStorage)
        api = mocker.Mock()

        def run(x):
            raise APIException("something broke")

        api.fetch_splits.side_effect = run
        storage.get_change_number.return_value = -1

        split_sync = SplitSynchronizer(api, storage)
        synchronizers = SplitSynchronizers(split_sync, mocker.Mock(), mocker.Mock(),
                                           mocker.Mock(), mocker.Mock(), mocker.Mock())

        synchronizer = Synchronizer(synchronizers, split_tasks)
        manager = Manager(threading.Event(), synchronizer,  mocker.Mock(), False)

        with pytest.raises(APIException):
            manager.start()

    def test_start_streaming_false(self, mocker):
        splits_ready_event = threading.Event()
        synchronizer = mocker.Mock(spec=Synchronizer)
        manager = Manager(splits_ready_event, synchronizer, mocker.Mock(), False)
        manager.start()

        splits_ready_event.wait(2)
        assert splits_ready_event.is_set()
        assert len(synchronizer.sync_all.mock_calls) == 1
        assert len(synchronizer.start_periodic_fetching.mock_calls) == 1
        assert len(synchronizer.start_periodic_data_recording.mock_calls) == 1



"""Splits API module."""

import logging
import json

from future.utils import raise_from

from splitio.api import APIException
from splitio.api.client import HttpClientException


_LOGGER = logging.getLogger(__name__)


class SplitsAPI(object):  # pylint: disable=too-few-public-methods
    """Class that uses an httpClient to communicate with the splits API."""

    def __init__(self, client, apikey):
        """
        Class constructor.

        :param client: HTTP Client responsble for issuing calls to the backend.
        :type client: HttpClient
        :param apikey: User apikey token.
        :type apikey: string
        """
        self._client = client
        self._apikey = apikey

    def fetch_splits(self, change_number):
        """
        Fetch splits from backend.

        :param changeNumber: Last known timestamp of a split modification.
        :type changeNumber: int

        :return: Json representation of a splitChanges response.
        :rtype: dict
        """
        try:
            response = self._client.get(
                'sdk',
                '/splitChanges',
                self._apikey,
                {'since': change_number}
            )
            if 200 <= response.status_code < 300:
                return json.loads(response.body)
            else:
                raise APIException(response.body, response.status_code)
        except HttpClientException as exc:
            _LOGGER.error('Error fetching splits because an exception was raised by the HTTPClient')
            _LOGGER.debug('Error: ', exc_info=True)
            raise_from(APIException('Splits not fetched correctly.'), exc)



"""UWSGI Cache Storage adapter module."""

from __future__ import absolute_import, division, print_function, unicode_literals
import time

try:
    #uwsgi is loaded at runtime by uwsgi app.
    import uwsgi
except ImportError:
    def missing_uwsgi_dependencies(*args, **kwargs):  #pylint: disable=unused-argument
        """Only complain for missing deps if they're used."""
        raise NotImplementedError('Missing uWSGI support dependencies.')
    uwsgi = missing_uwsgi_dependencies

# Cache used for locking & signaling keys
_SPLITIO_LOCK_CACHE_NAMESPACE = 'splitio_locks'

# Cache where split definitions are stored
_SPLITIO_SPLITS_CACHE_NAMESPACE = 'splitio_splits'

# Cache where segments are stored
_SPLITIO_SEGMENTS_CACHE_NAMESPACE = 'splitio_segments'

# Cache where impressions are stored
_SPLITIO_IMPRESSIONS_CACHE_NAMESPACE = 'splitio_impressions'

# Cache where metrics are stored
_SPLITIO_METRICS_CACHE_NAMESPACE = 'splitio_metrics'

# Cache where events are stored (1 key with lots of blocks)
_SPLITIO_EVENTS_CACHE_NAMESPACE = 'splitio_events'

# Cache where changeNumbers are stored
_SPLITIO_CHANGE_NUMBERS = 'splitio_changeNumbers'

# Cache with a big block size used for lists
_SPLITIO_MISC_NAMESPACE = 'splitio_misc'


class UWSGILock(object):
    """Context manager to be used for locking a key in the cache."""

    def __init__(self, adapter, key, overwrite_lock_seconds=5):
        """
        Initialize a lock with the key `key` and waits up to `overwrite_lock_seconds` to release.

        :param key: Key to be used.
        :type key: str

        :param overwrite_lock_seconds: How many seconds to wait before force-releasing.
        :type overwrite_lock_seconds: int
        """
        self._key = key
        self._overwrite_lock_seconds = overwrite_lock_seconds
        self._uwsgi = adapter

    def __enter__(self):
        """Loop until the lock is manually released or timeout occurs."""
        initial_time = time.time()
        while True:
            if not self._uwsgi.cache_exists(self._key, _SPLITIO_LOCK_CACHE_NAMESPACE):
                self._uwsgi.cache_set(self._key, str('locked'), 0, _SPLITIO_LOCK_CACHE_NAMESPACE)
                return
            else:
                if time.time() - initial_time > self._overwrite_lock_seconds:
                    return
            time.sleep(0.1)

    def __exit__(self, *args):
        """Remove lock."""
        self._uwsgi.cache_del(self._key, _SPLITIO_LOCK_CACHE_NAMESPACE)


class UWSGICacheEmulator(object):
    """UWSGI mock."""

    def __init__(self):
        """
        UWSGI Cache Emulator for unit tests. Implements uwsgi cache framework interface.

        http://uwsgi-docs.readthedocs.io/en/latest/Caching.html#accessing-the-cache-from-your-applications-using-the-cache-api
        """
        self._cache = dict()

    @staticmethod
    def _check_string_data_type(value):
        if type(value).__name__ == 'str':
            return True
        raise TypeError(
            'The value to add into uWSGI cache must be string and %s given' % type(value).__name__
        )

    def cache_get(self, key, cache_namespace='default'):
        """Get an element from cache."""
        if self.cache_exists(key, cache_namespace):
            return self._cache[cache_namespace][key]
        return None

    def cache_set(self, key, value, expires=0, cache_namespace='default'):  #pylint: disable=unused-argument
        """Set an elemen in the cache."""
        self._check_string_data_type(value)

        if cache_namespace in self._cache:
            self._cache[cache_namespace][key] = value
        else:
            self._cache[cache_namespace] = {key:value}

    def cache_update(self, key, value, expires=0, cache_namespace='default'):
        """Update an element."""
        self.cache_set(key, value, expires, cache_namespace)

    def cache_exists(self, key, cache_namespace='default'):
        """Return whether the element exists."""
        if cache_namespace in self._cache:
            if key in self._cache[cache_namespace]:
                return True
        return False

    def cache_del(self, key, cache_namespace='default'):
        """Delete an item from the cache."""
        if cache_namespace in self._cache:
            self._cache[cache_namespace].pop(key, None)

    def cache_clear(self, cache_namespace='default'):
        """Delete all elements in cache."""
        self._cache.pop(cache_namespace, None)


def get_uwsgi(emulator=False):
    """Return a uwsgi imported module or an emulator to use in unit test."""
    if emulator:
        return UWSGICacheEmulator()

    return uwsgi



"""A module for Split.io SDK API clients."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals


class Key(object):
    """Key class includes a matching key and bucketing key."""

    def __init__(self, matching_key, bucketing_key):
        """Construct a key object."""
        self._matching_key = matching_key
        self._bucketing_key = bucketing_key

    @property
    def matching_key(self):
        """Return matching key."""
        return self._matching_key

    @property
    def bucketing_key(self):
        """Return bucketing key."""
        return self._bucketing_key



"""Split Worker tests."""

import threading
import time
import pytest

from splitio.api.client import HttpResponse
from splitio.api import APIException
from splitio.storage import ImpressionStorage
from splitio.models.impressions import Impression
from splitio.sync.impression import ImpressionSynchronizer


class ImpressionsSynchronizerTests(object):
    """Impressions synchronizer test cases."""

    def test_synchronize_impressions_error(self, mocker):
        storage = mocker.Mock(spec=ImpressionStorage)
        storage.pop_many.return_value = [
            Impression('key1', 'split1', 'on', 'l1', 123456, 'b1', 321654),
            Impression('key2', 'split1', 'on', 'l1', 123456, 'b1', 321654),
        ]

        api = mocker.Mock()

        def run(x):
            raise APIException("something broke")
        api.flush_impressions.side_effect = run

        impression_synchronizer = ImpressionSynchronizer(api, storage, 5)
        impression_synchronizer.synchronize_impressions()
        assert impression_synchronizer._failed.qsize() == 2

    def test_synchronize_impressions_empty(self, mocker):
        storage = mocker.Mock(spec=ImpressionStorage)
        storage.pop_many.return_value = []

        api = mocker.Mock()

        def run(x):
            run._called += 1

        run._called = 0
        api.flush_impressions.side_effect = run
        impression_synchronizer = ImpressionSynchronizer(api, storage, 5)
        impression_synchronizer.synchronize_impressions()
        assert run._called == 0

    def test_synchronize_impressions(self, mocker):
        storage = mocker.Mock(spec=ImpressionStorage)
        storage.pop_many.return_value = [
            Impression('key1', 'split1', 'on', 'l1', 123456, 'b1', 321654),
            Impression('key2', 'split1', 'on', 'l1', 123456, 'b1', 321654),
        ]

        api = mocker.Mock()

        def run(x):
            run._called += 1
            return HttpResponse(200, '')

        api.flush_impressions.side_effect = run
        run._called = 0

        impression_synchronizer = ImpressionSynchronizer(api, storage, 5)
        impression_synchronizer.synchronize_impressions()
        assert run._called == 1
        assert impression_synchronizer._failed.qsize() == 0



"""Base storage interfaces."""
from __future__ import absolute_import

import abc

from six import add_metaclass


@add_metaclass(abc.ABCMeta)
class SplitStorage(object):
    """Split storage interface implemented as an abstract class."""

    @abc.abstractmethod
    def get(self, split_name):
        """
        Retrieve a split.

        :param split_name: Name of the feature to fetch.
        :type split_name: str

        :rtype: str
        """
        pass

    @abc.abstractmethod
    def fetch_many(self, split_names):
        """
        Retrieve splits.

        :param split_names: Names of the features to fetch.
        :type split_names: list(str)

        :rtype: dict
        """
        pass

    @abc.abstractmethod
    def put(self, split):
        """
        Store a split.

        :param split: Split object to store
        :type split_name: splitio.models.splits.Split
        """
        pass

    @abc.abstractmethod
    def remove(self, split_name):
        """
        Remove a split from storage.

        :param split_name: Name of the feature to remove.
        :type split_name: str

        :return: True if the split was found and removed. False otherwise.
        :rtype: bool
        """
        pass

    @abc.abstractmethod
    def get_change_number(self):
        """
        Retrieve latest split change number.

        :rtype: int
        """
        pass

    @abc.abstractmethod
    def set_change_number(self, new_change_number):
        """
        Set the latest change number.

        :param new_change_number: New change number.
        :type new_change_number: int
        """
        pass

    @abc.abstractmethod
    def get_split_names(self):
        """
        Retrieve a list of all split names.

        :return: List of split names.
        :rtype: list(str)
        """
        pass

    @abc.abstractmethod
    def get_all_splits(self):
        """
        Return all the splits.

        :return: List of all the splits.
        :rtype: list
        """
        pass

    @abc.abstractmethod
    def is_valid_traffic_type(self, traffic_type_name):
        """
        Return whether the traffic type exists in at least one split in cache.

        :param traffic_type_name: Traffic type to validate.
        :type traffic_type_name: str

        :return: True if the traffic type is valid. False otherwise.
        :rtype: bool
        """
        pass

    def get_segment_names(self):
        """
        Return a set of all segments referenced by splits in storage.

        :return: Set of all segment names.
        :rtype: set(string)
        """
        return set([name for spl in self.get_all_splits() for name in spl.get_segment_names()])

    @abc.abstractmethod
    def kill_locally(self, split_name, default_treatment, change_number):
        """
        Local kill for split

        :param split_name: name of the split to perform kill
        :type split_name: str
        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        pass


@add_metaclass(abc.ABCMeta)
class SegmentStorage(object):
    """Segment storage interface implemented as an abstract class."""

    @abc.abstractmethod
    def get(self, segment_name):
        """
        Retrieve a segment.

        :param segment_name: Name of the segment to fetch.
        :type segment_name: str

        :rtype: str
        """
        pass

    @abc.abstractmethod
    def put(self, segment):
        """
        Store a segment.

        :param segment: Segment to store.
        :type segment: splitio.models.segment.Segment
        """
        pass

    @abc.abstractmethod
    def update(self, segment_name, to_add, to_remove, change_number=None):
        """
        Store a split.

        :param segment_name: Name of the segment to update.
        :type segment_name: str
        :param to_add: List of members to add to the segment.
        :type to_add: list
        :param to_remove: List of members to remove from the segment.
        :type to_remove: list
        """
        pass

    @abc.abstractmethod
    def get_change_number(self, segment_name):
        """
        Retrieve latest change number for a segment.

        :param segment_name: Name of the segment.
        :type segment_name: str

        :rtype: int
        """
        pass

    @abc.abstractmethod
    def set_change_number(self, segment_name, new_change_number):
        """
        Set the latest change number.

        :param segment_name: Name of the segment.
        :type segment_name: str
        :param new_change_number: New change number.
        :type new_change_number: int
        """
        pass

    @abc.abstractmethod
    def segment_contains(self, segment_name, key):
        """
        Check whether a specific key belongs to a segment in storage.

        :param segment_name: Name of the segment to search in.
        :type segment_name: str
        :param key: Key to search for.
        :type key: str

        :return: True if the segment contains the key. False otherwise.
        :rtype: bool
        """
        pass


@add_metaclass(abc.ABCMeta)
class ImpressionStorage(object):
    """Impressions storage interface."""

    @abc.abstractmethod
    def put(self, impressions):
        """
        Put one or more impressions in storage.

        :param impressions: List of one or more impressions to store.
        :type impressions: list
        """
        pass

    @abc.abstractmethod
    def pop_many(self, count):
        """
        Pop the oldest N impressions from storage.

        :param count: Number of impressions to pop.
        :type count: int
        """
        pass


@add_metaclass(abc.ABCMeta)
class EventStorage(object):
    """Events storage interface."""

    @abc.abstractmethod
    def put(self, events):
        """
        Put one or more events in storage.

        :param events: List of one or more events to store.
        :type events: list
        """
        pass

    @abc.abstractmethod
    def pop_many(self, count):
        """
        Pop the oldest N events from storage.

        :param count: Number of events to pop.
        :type count: int
        """
        pass


@add_metaclass(abc.ABCMeta)
class TelemetryStorage(object):
    """Telemetry storage interface."""

    @abc.abstractmethod
    def inc_latency(self, name, bucket):
        """
        Add a latency.

        :param name: Name of the latency metric.
        :type name: str
        :param value: Value of the latency metric.
        :tyoe value: int
        """
        pass

    @abc.abstractmethod
    def inc_counter(self, name):
        """
        Increment a counter.

        :param name: Name of the counter metric.
        :type name: str
        """
        pass

    @abc.abstractmethod
    def put_gauge(self, name, value):
        """
        Add a gauge metric.

        :param name: Name of the gauge metric.
        :type name: str
        :param value: Value of the gauge metric.
        :type value: int
        """
        pass

    @abc.abstractmethod
    def pop_counters(self):
        """
        Get all the counters.

        :rtype: list
        """
        pass

    @abc.abstractmethod
    def pop_gauges(self):
        """
        Get all the gauges.

        :rtype: list

        """
        pass

    @abc.abstractmethod
    def pop_latencies(self):
        """
        Get all latencies.

        :rtype: list
        """
        pass



"""A module for Split.io Managers."""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from . import input_validator


_LOGGER = logging.getLogger(__name__)


class SplitManager(object):
    """Split Manager. Gives insights on data cached by splits."""

    def __init__(self, factory):
        """
        Class constructor.

        :param factory: Factory containing all storage references.
        :type factory: splitio.client.factory.SplitFactory
        """
        self._factory = factory
        self._storage = factory._get_storage('splits')  # pylint: disable=protected-access

    def split_names(self):
        """
        Get the name of fetched splits.

        :return: A list of str
        :rtype: list
        """
        if self._factory.destroyed:
            _LOGGER.error("Client has already been destroyed - no calls possible.")
            return []

        if not self._factory.ready:
            _LOGGER.warning(
                "split_names: The SDK is not ready, results may be incorrect. "
                "Make sure to wait for SDK readiness before using this method"
            )

        return self._storage.get_split_names()

    def splits(self):
        """
        Get the fetched splits. Subclasses need to override this method.

        :return: A List of SplitView.
        :rtype: list()
        """
        if self._factory.destroyed:
            _LOGGER.error("Client has already been destroyed - no calls possible.")
            return []

        if not self._factory.ready:
            _LOGGER.warning(
                "splits: The SDK is not ready, results may be incorrect. "
                "Make sure to wait for SDK readiness before using this method"
            )

        return [split.to_split_view() for split in self._storage.get_all_splits()]

    def split(self, feature_name):
        """
        Get the splitView of feature_name. Subclasses need to override this method.

        :param feature_name: Name of the feture to retrieve.
        :type feature_name: str

        :return: The SplitView instance.
        :rtype: splitio.models.splits.SplitView
        """
        if self._factory.destroyed:
            _LOGGER.error("Client has already been destroyed - no calls possible.")
            return []

        feature_name = input_validator.validate_manager_feature_name(
            feature_name,
            self._factory.ready,
            self._storage
        )

        if not self._factory.ready:
            _LOGGER.warning(
                "split: The SDK is not ready, results may be incorrect. "
                "Make sure to wait for SDK readiness before using this method"
            )

        if feature_name is None:
            return None

        split = self._storage.get(feature_name)
        return split.to_split_view() if split is not None else None



"""Message processor tests."""
from queue import Queue
from splitio.push.processor import MessageProcessor
from splitio.sync.synchronizer import Synchronizer
from splitio.push.parser import SplitChangeUpdate, SegmentChangeUpdate, SplitKillUpdate


class ProcessorTests(object):
    """Message processor test cases."""

    def test_split_change(self, mocker):
        """Test split change is properly handled."""
        sync_mock = mocker.Mock(spec=Synchronizer)
        queue_mock = mocker.Mock(spec=Queue)
        mocker.patch('splitio.push.processor.Queue', new=queue_mock)
        processor = MessageProcessor(sync_mock)
        update = SplitChangeUpdate('sarasa', 123, 123)
        processor.handle(update)
        assert queue_mock.mock_calls == [
            mocker.call(),  # construction of split queue
            mocker.call(),  # construction of split queue
            mocker.call().put(update)
        ]

    def test_split_kill(self, mocker):
        """Test split kill is properly handled."""
        sync_mock = mocker.Mock(spec=Synchronizer)
        queue_mock = mocker.Mock(spec=Queue)
        mocker.patch('splitio.push.processor.Queue', new=queue_mock)
        processor = MessageProcessor(sync_mock)
        update = SplitKillUpdate('sarasa', 123, 456, 'some_split', 'off')
        processor.handle(update)
        assert queue_mock.mock_calls == [
            mocker.call(),  # construction of split queue
            mocker.call(),  # construction of split queue
            mocker.call().put(update)
        ]
        assert sync_mock.kill_split.mock_calls == [
            mocker.call('some_split', 'off', 456)
        ]

    def test_segment_change(self, mocker):
        """Test segment change is properly handled."""
        sync_mock = mocker.Mock(spec=Synchronizer)
        queue_mock = mocker.Mock(spec=Queue)
        mocker.patch('splitio.push.processor.Queue', new=queue_mock)
        processor = MessageProcessor(sync_mock)
        update = SegmentChangeUpdate('sarasa', 123, 123, 'some_segment')
        processor.handle(update)
        assert queue_mock.mock_calls == [
            mocker.call(),  # construction of split queue
            mocker.call(),  # construction of split queue
            mocker.call().put(update)
        ]

    def test_todo(self):
        """Fix previous tests so that we validate WHICH queue the update is pushed into."""
        assert NotImplementedError("DO THAT")



"""This module contains everything related to split.io exceptions"""
from __future__ import absolute_import, division, print_function, unicode_literals

from splitio.client.factory import TimeoutException
from splitio.storage.adapters.redis import SentinelConfigurationException



"""A module for Split.io Factories."""
from __future__ import absolute_import, division, print_function, unicode_literals


import logging
import threading
from collections import Counter

from enum import Enum

from splitio.client.client import Client
from splitio.client import input_validator
from splitio.client.manager import SplitManager
from splitio.client.config import sanitize as sanitize_config
from splitio.client import util
from splitio.client.listener import ImpressionListenerWrapper
from splitio.engine.impressions import Manager as ImpressionsManager

# Storage
from splitio.storage.inmemmory import InMemorySplitStorage, InMemorySegmentStorage, \
    InMemoryImpressionStorage, InMemoryEventStorage, InMemoryTelemetryStorage
from splitio.storage.adapters import redis
from splitio.storage.redis import RedisSplitStorage, RedisSegmentStorage, RedisImpressionsStorage, \
    RedisEventsStorage, RedisTelemetryStorage
from splitio.storage.adapters.uwsgi_cache import get_uwsgi
from splitio.storage.uwsgi import UWSGIEventStorage, UWSGIImpressionStorage, UWSGISegmentStorage, \
    UWSGISplitStorage, UWSGITelemetryStorage

# APIs
from splitio.api.client import HttpClient
from splitio.api.splits import SplitsAPI
from splitio.api.segments import SegmentsAPI
from splitio.api.impressions import ImpressionsAPI
from splitio.api.events import EventsAPI
from splitio.api.telemetry import TelemetryAPI
from splitio.api.auth import AuthAPI


# Tasks
from splitio.tasks.split_sync import SplitSynchronizationTask
from splitio.tasks.segment_sync import SegmentSynchronizationTask
from splitio.tasks.impressions_sync import ImpressionsSyncTask, ImpressionsCountSyncTask
from splitio.tasks.events_sync import EventsSyncTask
from splitio.tasks.telemetry_sync import TelemetrySynchronizationTask

# Synchronizer
from splitio.sync.synchronizer import SplitTasks, SplitSynchronizers, Synchronizer, \
    LocalhostSynchronizer
from splitio.sync.manager import Manager
from splitio.sync.split import SplitSynchronizer, LocalSplitSynchronizer
from splitio.sync.segment import SegmentSynchronizer
from splitio.sync.impression import ImpressionSynchronizer, ImpressionsCountSynchronizer
from splitio.sync.event import EventSynchronizer
from splitio.sync.telemetry import TelemetrySynchronizer

# Localhost stuff
from splitio.client.localhost import LocalhostEventsStorage, LocalhostImpressionsStorage, \
    LocalhostTelemetryStorage


_LOGGER = logging.getLogger(__name__)
_INSTANTIATED_FACTORIES = Counter()
_INSTANTIATED_FACTORIES_LOCK = threading.RLock()


class Status(Enum):
    """Factory Status."""

    NOT_INITIALIZED = 'NOT_INITIALIZED'
    READY = 'READY'
    DESTROYED = 'DESTROYED'


class TimeoutException(Exception):
    """Exception to be raised upon a block_until_ready call when a timeout expires."""

    pass


class SplitFactory(object):  # pylint: disable=too-many-instance-attributes
    """Split Factory/Container class."""

    def __init__(  # pylint: disable=too-many-arguments
            self,
            apikey,
            storages,
            labels_enabled,
            impressions_manager,
            sync_manager=None,
            sdk_ready_flag=None,
    ):
        """
        Class constructor.

        :param storages: Dictionary of storages for all split models.
        :type storages: dict
        :param labels_enabled: Whether the impressions should store labels or not.
        :type labels_enabled: bool
        :param apis: Dictionary of apis client wrappers
        :type apis: dict
        :param sync_manager: Manager synchronization
        :type sync_manager: splitio.sync.manager.Manager
        :param sdk_ready_flag: Event to set when the sdk is ready.
        :type sdk_ready_flag: threading.Event
        :param impression_manager: Impressions manager instance
        :type impression_listener: ImpressionsManager
        """
        self._apikey = apikey
        self._storages = storages
        self._labels_enabled = labels_enabled
        self._sync_manager = sync_manager
        self._sdk_internal_ready_flag = sdk_ready_flag
        self._sdk_ready_flag = threading.Event()
        self._impressions_manager = impressions_manager

        # If we have a ready flag, it means we have sync tasks that need to finish
        # before the SDK client becomes ready.
        if self._sdk_internal_ready_flag is not None:
            self._status = Status.NOT_INITIALIZED
            # add a listener that updates the status to READY once the flag is set.
            ready_updater = threading.Thread(target=self._update_status_when_ready,
                                             name='SDKReadyFlagUpdater')
            ready_updater.setDaemon(True)
            ready_updater.start()
        else:
            self._status = Status.READY

    def _update_status_when_ready(self):
        """Wait until the sdk is ready and update the status."""
        self._sdk_internal_ready_flag.wait()
        self._status = Status.READY
        self._sdk_ready_flag.set()

    def _get_storage(self, name):
        """
        Return a reference to the specified storage.

        :param name: Name of the requested storage.
        :type name: str

        :return: requested factory.
        :rtype: object
        """
        return self._storages[name]

    def client(self):
        """
        Return a new client.

        This client is only a set of references to structures hold by the factory.
        Creating one a fast operation and safe to be used anywhere.
        """
        return Client(self, self._impressions_manager, self._labels_enabled)

    def manager(self):
        """
        Return a new manager.

        This manager is only a set of references to structures hold by the factory.
        Creating one a fast operation and safe to be used anywhere.
        """
        return SplitManager(self)

    def block_until_ready(self, timeout=None):
        """
        Blocks until the sdk is ready or the timeout specified by the user expires.

        When ready, the factory's status is updated accordingly.

        :param timeout: Number of seconds to wait (fractions allowed)
        :type timeout: int
        """
        if self._sdk_internal_ready_flag is not None:
            ready = self._sdk_ready_flag.wait(timeout)

            if not ready:
                raise TimeoutException('SDK Initialization: time of %d exceeded' % timeout)

    @property
    def ready(self):
        """
        Return whether the factory is ready.

        :return: True if the factory is ready. False otherwhise.
        :rtype: bool
        """
        return self._status == Status.READY

    def destroy(self, destroyed_event=None):
        """
        Destroy the factory and render clients unusable.

        Destroy frees up storage taken but split data, flushes impressions & events,
        and invalidates the clients, making them return control.

        :param destroyed_event: Event to signal when destroy process has finished.
        :type destroyed_event: threading.Event
        """
        if self.destroyed:
            _LOGGER.info('Factory already destroyed.')
            return

        try:
            if self._sync_manager is not None:
                if destroyed_event is not None:

                    def _wait_for_tasks_to_stop():
                        self._sync_manager.stop(True)
                        destroyed_event.set()

                    wait_thread = threading.Thread(target=_wait_for_tasks_to_stop)
                    wait_thread.setDaemon(True)
                    wait_thread.start()
                else:
                    self._sync_manager.stop(False)
            elif destroyed_event is not None:
                destroyed_event.set()
        finally:
            self._status = Status.DESTROYED
            with _INSTANTIATED_FACTORIES_LOCK:
                _INSTANTIATED_FACTORIES.subtract([self._apikey])

    @property
    def destroyed(self):
        """
        Return whether the factory has been destroyed or not.

        :return: True if the factory has been destroyed. False otherwise.
        :rtype: bool
        """
        return self._status == Status.DESTROYED


def _wrap_impression_listener(listener, metadata):
    """
    Wrap the impression listener if any.

    :param listener: User supplied impression listener or None
    :type listener: splitio.client.listener.ImpressionListener | None
    :param metadata: SDK Metadata
    :type metadata: splitio.client.util.SdkMetadata
    """
    if listener is not None:
        return ImpressionListenerWrapper(listener, metadata)
    return None


def _build_in_memory_factory(api_key, cfg, sdk_url=None, events_url=None,  # pylint:disable=too-many-arguments,too-many-locals
                             auth_api_base_url=None, streaming_api_base_url=None):
    """Build and return a split factory tailored to the supplied config."""
    if not input_validator.validate_factory_instantiation(api_key):
        return None

    http_client = HttpClient(
        sdk_url=sdk_url,
        events_url=events_url,
        auth_url=auth_api_base_url,
        timeout=cfg.get('connectionTimeout')
    )

    sdk_metadata = util.get_metadata(cfg)
    apis = {
        'auth': AuthAPI(http_client, api_key, sdk_metadata),
        'splits': SplitsAPI(http_client, api_key),
        'segments': SegmentsAPI(http_client, api_key),
        'impressions': ImpressionsAPI(http_client, api_key, sdk_metadata, cfg['impressionsMode']),
        'events': EventsAPI(http_client, api_key, sdk_metadata),
        'telemetry': TelemetryAPI(http_client, api_key, sdk_metadata)
    }

    if not input_validator.validate_apikey_type(apis['segments']):
        return None

    storages = {
        'splits': InMemorySplitStorage(),
        'segments': InMemorySegmentStorage(),
        'impressions': InMemoryImpressionStorage(cfg['impressionsQueueSize']),
        'events': InMemoryEventStorage(cfg['eventsQueueSize']),
        'telemetry': InMemoryTelemetryStorage()
    }

    imp_manager = ImpressionsManager(
        storages['impressions'].put,
        cfg['impressionsMode'],
        True,
        _wrap_impression_listener(cfg['impressionListener'], sdk_metadata))

    synchronizers = SplitSynchronizers(
        SplitSynchronizer(apis['splits'], storages['splits']),
        SegmentSynchronizer(apis['segments'], storages['splits'], storages['segments']),
        ImpressionSynchronizer(apis['impressions'], storages['impressions'],
                               cfg['impressionsBulkSize']),
        EventSynchronizer(apis['events'], storages['events'], cfg['eventsBulkSize']),
        TelemetrySynchronizer(apis['telemetry'], storages['telemetry']),
        ImpressionsCountSynchronizer(apis['impressions'], imp_manager),
    )

    tasks = SplitTasks(
        SplitSynchronizationTask(
            synchronizers.split_sync.synchronize_splits,
            cfg['featuresRefreshRate'],
        ),
        SegmentSynchronizationTask(
            synchronizers.segment_sync.synchronize_segments,
            cfg['segmentsRefreshRate'],
        ),
        ImpressionsSyncTask(
            synchronizers.impressions_sync.synchronize_impressions,
            cfg['impressionsRefreshRate'],
        ),
        EventsSyncTask(synchronizers.events_sync.synchronize_events, cfg['eventsPushRate']),
        TelemetrySynchronizationTask(
            synchronizers.telemetry_sync.synchronize_telemetry,
            cfg['metricsRefreshRate'],
        ),
        ImpressionsCountSyncTask(synchronizers.impressions_count_sync.synchronize_counters)
    )

    synchronizer = Synchronizer(synchronizers, tasks)

    sdk_ready_flag = threading.Event()
    manager = Manager(sdk_ready_flag, synchronizer, apis['auth'], cfg['streamingEnabled'],
                      streaming_api_base_url)

    initialization_thread = threading.Thread(target=manager.start, name="SDKInitializer")
    initialization_thread.setDaemon(True)
    initialization_thread.start()

    storages['events'].set_queue_full_hook(tasks.events_task.flush)
    storages['impressions'].set_queue_full_hook(tasks.impressions_task.flush)

    return SplitFactory(api_key, storages, cfg['labelsEnabled'],
                        imp_manager, manager, sdk_ready_flag)


def _build_redis_factory(api_key, cfg):
    """Build and return a split factory with redis-based storage."""
    sdk_metadata = util.get_metadata(cfg)
    redis_adapter = redis.build(cfg)
    cache_enabled = cfg.get('redisLocalCacheEnabled', False)
    cache_ttl = cfg.get('redisLocalCacheTTL', 5)
    storages = {
        'splits': RedisSplitStorage(redis_adapter, cache_enabled, cache_ttl),
        'segments': RedisSegmentStorage(redis_adapter),
        'impressions': RedisImpressionsStorage(redis_adapter, sdk_metadata),
        'events': RedisEventsStorage(redis_adapter, sdk_metadata),
        'telemetry': RedisTelemetryStorage(redis_adapter, sdk_metadata)
    }
    return SplitFactory(
        api_key,
        storages,
        cfg['labelsEnabled'],
        ImpressionsManager(storages['impressions'].put, cfg['impressionsMode'], False,
                           _wrap_impression_listener(cfg['impressionListener'], sdk_metadata))
    )


def _build_uwsgi_factory(api_key, cfg):
    """Build and return a split factory with redis-based storage."""
    sdk_metadata = util.get_metadata(cfg)
    uwsgi_adapter = get_uwsgi()
    storages = {
        'splits': UWSGISplitStorage(uwsgi_adapter),
        'segments': UWSGISegmentStorage(uwsgi_adapter),
        'impressions': UWSGIImpressionStorage(uwsgi_adapter),
        'events': UWSGIEventStorage(uwsgi_adapter),
        'telemetry': UWSGITelemetryStorage(uwsgi_adapter)
    }
    return SplitFactory(
        api_key,
        storages,
        cfg['labelsEnabled'],
        ImpressionsManager(storages['impressions'].put, cfg['impressionsMode'], True,
                           _wrap_impression_listener(cfg['impressionListener'], sdk_metadata))
    )


def _build_localhost_factory(cfg):
    """Build and return a localhost factory for testing/development purposes."""
    storages = {
        'splits': InMemorySplitStorage(),
        'segments': InMemorySegmentStorage(),  # not used, just to avoid possible future errors.
        'impressions': LocalhostImpressionsStorage(),
        'events': LocalhostEventsStorage(),
        'telemetry': LocalhostTelemetryStorage()
    }

    synchronizers = SplitSynchronizers(
        LocalSplitSynchronizer(cfg['splitFile'], storages['splits']),
        None, None, None, None, None,
    )

    tasks = SplitTasks(
        SplitSynchronizationTask(
            synchronizers.split_sync.synchronize_splits,
            cfg['featuresRefreshRate'],
        ), None, None, None, None, None,
    )

    ready_event = threading.Event()
    synchronizer = LocalhostSynchronizer(synchronizers, tasks)
    manager = Manager(ready_event, synchronizer, None, False)
    manager.start()

    return SplitFactory(
        'localhost',
        storages,
        False,
        ImpressionsManager(storages['impressions'].put, cfg['impressionsMode'], True, None),
        manager,
        ready_event
    )


def get_factory(api_key, **kwargs):
    """Build and return the appropriate factory."""
    try:
        _INSTANTIATED_FACTORIES_LOCK.acquire()
        if _INSTANTIATED_FACTORIES:
            if api_key in _INSTANTIATED_FACTORIES:
                _LOGGER.warning(
                    "factory instantiation: You already have %d %s with this API Key. "
                    "We recommend keeping only one instance of the factory at all times "
                    "(Singleton pattern) and reusing it throughout your application.",
                    _INSTANTIATED_FACTORIES[api_key],
                    'factory' if _INSTANTIATED_FACTORIES[api_key] == 1 else 'factories'
                )
            else:
                _LOGGER.warning(
                    "factory instantiation: You already have an instance of the Split factory. "
                    "Make sure you definitely want this additional instance. "
                    "We recommend keeping only one instance of the factory at all times "
                    "(Singleton pattern) and reusing it throughout your application."
                )

        config = sanitize_config(api_key, kwargs.get('config', {}))

        if config['operationMode'] == 'localhost-standalone':
            return _build_localhost_factory(config)

        if config['operationMode'] == 'redis-consumer':
            return _build_redis_factory(api_key, config)

        if config['operationMode'] == 'uwsgi-consumer':
            return _build_uwsgi_factory(api_key, config)

        return _build_in_memory_factory(
            api_key,
            config,
            kwargs.get('sdk_api_base_url'),
            kwargs.get('events_api_base_url'),
            kwargs.get('auth_api_base_url'),
            kwargs.get('streaming_api_base_url')
        )
    finally:
        _INSTANTIATED_FACTORIES.update([api_key])
        _INSTANTIATED_FACTORIES_LOCK.release()



"""Default settings for the Split.IO SDK Python client."""
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path
import logging

from splitio.engine.impressions import ImpressionsMode


_LOGGER = logging.getLogger(__name__)


DEFAULT_CONFIG = {
    'operationMode': 'in-memory',
    'connectionTimeout': 1500,
    'splitSdkMachineName': None,
    'splitSdkMachineIp': None,
    'streamingEnabled': True,
    'featuresRefreshRate': 30,
    'segmentsRefreshRate': 30,
    'metricsRefreshRate': 60,
    'impressionsRefreshRate': 5 * 60,
    'impressionsBulkSize': 5000,
    'impressionsQueueSize': 10000,
    'eventsPushRate': 10,
    'eventsBulkSize': 5000,
    'eventsQueueSize': 10000,
    'labelsEnabled': True,
    'IPAddressesEnabled': True,
    'impressionsMode': 'OPTIMIZED',
    'impressionListener': None,
    'redisLocalCacheEnabled': False,
    'redisLocalCacheTTL': 5,
    'redisHost': 'localhost',
    'redisPort': 6379,
    'redisDb': 0,
    'redisPassword': None,
    'redisSocketTimeout': None,
    'redisSocketConnectTimeout': None,
    'redisSocketKeepalive': None,
    'redisSocketKeepaliveOptions': None,
    'redisConnectionPool': None,
    'redisUnixSocketPath': None,
    'redisEncoding': 'utf-8',
    'redisEncodingErrors': 'strict',
    'redisCharset': None,
    'redisErrors': None,
    'redisDecodeResponses': False,
    'redisRetryOnTimeout': False,
    'redisSsl': False,
    'redisSslKeyfile': None,
    'redisSslCertfile': None,
    'redisSslCertReqs': None,
    'redisSslCaCerts': None,
    'redisMaxConnections': None,
    'machineName': None,
    'machineIp': None,
    'splitFile': os.path.join(os.path.expanduser('~'), '.split')
}


def _parse_operation_mode(apikey, config):
    """
    Process incoming config to determine operation mode.

    :param config: user supplied config
    :type config: dict

    :returns: operation mode
    :rtype: str
    """
    if apikey == 'localhost':
        return 'localhost-standalone'

    if 'redisHost' in config or 'redisSentinels' in config:
        return 'redis-consumer'

    if 'uwsgiClient' in config:
        return 'uwsgi-consumer'

    return 'inmemory-standalone'


def _sanitize_impressions_mode(mode, refresh_rate=None):
    """
    Check supplied impressions mode and adjust refresh rate.

    :param config: default + supplied config
    :type config: dict

    :returns: config with sanitized impressions mode & refresh rate
    :rtype: config
    """
    if not isinstance(mode, ImpressionsMode):
        try:
            mode = ImpressionsMode(mode.upper())
        except (ValueError, AttributeError):
            _LOGGER.warning('You passed an invalid impressionsMode, impressionsMode should be '
                            'one of the following values: `debug` or `optimized`. '
                            'Defaulting to `optimized` mode.')
            mode = ImpressionsMode.OPTIMIZED

    if mode == ImpressionsMode.DEBUG:
        refresh_rate = max(1, refresh_rate) if refresh_rate is not None else 60
    else:
        refresh_rate = max(60, refresh_rate) if refresh_rate is not None else 5 * 60

    return mode, refresh_rate


def sanitize(apikey, config):
    """
    Look for inconsistencies or ill-formed configs and tune it accordingly.

    :param apikey: customer's apikey
    :type apikey: str

    :param config: DEFAULT + user supplied config
    :type config: dict

    :returns: sanitized config
    :rtype: dict
    """
    config['operationMode'] = _parse_operation_mode(apikey, config)
    processed = DEFAULT_CONFIG.copy()
    processed.update(config)
    imp_mode, imp_rate = _sanitize_impressions_mode(config.get('impressionsMode'),
                                                    config.get('impressionsRefreshRate'))
    processed['impressionsMode'] = imp_mode
    processed['impressionsRefreshRate'] = imp_rate
    return processed



"""Message processor & Notification manager keeper implementations."""

from queue import Queue

from six import raise_from

from splitio.push.parser import UpdateType
from splitio.push.splitworker import SplitWorker
from splitio.push.segmentworker import SegmentWorker


class MessageProcessor(object):
    """Message processor class."""

    def __init__(self, synchronizer):
        """
        Class constructor.

        :param synchronizer: synchronizer component
        :type synchronizer: splitio.sync.synchronizer.Synchronizer
        """
        self._split_queue = Queue()
        self._segments_queue = Queue()
        self._synchronizer = synchronizer
        self._split_worker = SplitWorker(synchronizer.synchronize_splits, self._split_queue)
        self._segments_worker = SegmentWorker(synchronizer.synchronize_segment, self._segments_queue)
        self._handlers = {
            UpdateType.SPLIT_UPDATE: self._handle_split_update,
            UpdateType.SPLIT_KILL: self._handle_split_kill,
            UpdateType.SEGMENT_UPDATE: self._handle_segment_change
        }

    def _handle_split_update(self, event):
        """
        Handle incoming split update notification.

        :param event: Incoming split change event
        :type event: splitio.push.parser.SplitChangeUpdate
        """
        self._split_queue.put(event)

    def _handle_split_kill(self, event):
        """
        Handle incoming split kill notification.

        :param event: Incoming split kill event
        :type event: splitio.push.parser.SplitKillUpdate
        """
        self._synchronizer.kill_split(event.split_name, event.default_treatment,
                                      event.change_number)
        self._split_queue.put(event)

    def _handle_segment_change(self, event):
        """
        Handle incoming segment update notification.

        :param event: Incoming segment change event
        :type event: splitio.push.parser.Update
        """
        self._segments_queue.put(event)

    def update_workers_status(self, enabled):
        """
        Enable/Disable push update workers.

        :param enabled: if True, enable workers. If False, disable them.
        :type enabled: bool
        """
        if enabled:
            self._split_worker.start()
            self._segments_worker.start()
        else:
            self._split_worker.stop()
            self._segments_worker.stop()

    def handle(self, event):
        """
        Handle incoming update event.

        :param event: incoming data update event.
        :type event: splitio.push.BaseUpdate
        """
        try:
            handle = self._handlers[event.update_type]
        except KeyError as exc:
            raise_from('no handler for notification type: %s' % event.update_type, exc)

        handle(event)

    def shutdown(self):
        """Stop splits & segments workers."""
        self._split_worker.stop()
        self._segments_worker.stop()



"""SDK Telemetry helpers."""
from bisect import bisect_left


BUCKETS = (
    1000, 1500, 2250, 3375, 5063,
    7594, 11391, 17086, 25629, 38443,
    57665, 86498, 129746, 194620, 291929,
    437894, 656841, 985261, 1477892, 2216838,
    3325257, 4987885, 7481828
)
MAX_LATENCY = 7481828


def get_latency_bucket_index(micros):
    """
    Find the bucket index for a measured latency.

    :param micros: Measured latency in microseconds
    :type micros: int
    :return: Bucket index for the given latency
    :rtype: int
    """
    if micros > MAX_LATENCY:
        return len(BUCKETS) - 1

    return bisect_left(BUCKETS, micros)



"""UWSGI Cache based storages implementation module."""
import logging
import json

from splitio.storage import SplitStorage, SegmentStorage, ImpressionStorage, EventStorage, \
    TelemetryStorage
from splitio.models import splits, segments
from splitio.models.impressions import Impression
from splitio.models.events import Event
from splitio.storage.adapters.uwsgi_cache import _SPLITIO_CHANGE_NUMBERS, \
    _SPLITIO_EVENTS_CACHE_NAMESPACE, _SPLITIO_IMPRESSIONS_CACHE_NAMESPACE, \
    _SPLITIO_METRICS_CACHE_NAMESPACE, _SPLITIO_MISC_NAMESPACE, UWSGILock, \
    _SPLITIO_SEGMENTS_CACHE_NAMESPACE, _SPLITIO_SPLITS_CACHE_NAMESPACE, \
    _SPLITIO_LOCK_CACHE_NAMESPACE


_LOGGER = logging.getLogger(__name__)


class UWSGISplitStorage(SplitStorage):
    """UWSGI-Cache based implementation of a split storage."""

    _KEY_TEMPLATE = 'split.{suffix}'
    _KEY_TILL = 'splits.till'
    _KEY_FEATURE_LIST = 'splits.list'
    _KEY_FEATURE_LIST_LOCK = 'splits.list.lock'
    _KEY_TRAFFIC_TYPES = 'splits.traffic_types'
    _KEY_TRAFFIC_TYPES_LOCK = 'splits.traffic_types.lock'
    _OVERWRITE_LOCK_SECONDS = 5

    def __init__(self, uwsgi_entrypoint):
        """
        Class constructor.

        :param uwsgi_entrypoint: UWSGI module. Can be the actual module or a mock.
        :type uwsgi_entrypoint: module
        """
        self._uwsgi = uwsgi_entrypoint

    def get(self, split_name):
        """
        Retrieve a split.

        :param split_name: Name of the feature to fetch.
        :type split_name: str

        :rtype: str
        """
        raw = self._uwsgi.cache_get(
            self._KEY_TEMPLATE.format(suffix=split_name),
            _SPLITIO_SPLITS_CACHE_NAMESPACE
        )
        to_return = splits.from_raw(json.loads(raw)) if raw is not None else None
        if not to_return:
            _LOGGER.warning("Trying to retrieve nonexistant split %s. Ignoring.", split_name)
        return to_return

    def fetch_many(self, split_names):
        """
        Retrieve splits.

        :param split_names: Names of the features to fetch.
        :type split_name: list(str)

        :return: A dict with split objects parsed from queue.
        :rtype: dict(split_name, splitio.models.splits.Split)
        """
        return {split_name: self.get(split_name) for split_name in split_names}

    def put(self, split):
        """
        Store a split.

        :param split: Split object to store
        :type split: splitio.models.splits.Split
        """
        self._uwsgi.cache_update(
            self._KEY_TEMPLATE.format(suffix=split.name),
            json.dumps(split.to_json()),
            0,
            _SPLITIO_SPLITS_CACHE_NAMESPACE
        )
        self._add_split_to_list(split.name)
        self._increase_traffic_type_count(split.traffic_type_name)

    def remove(self, split_name):
        """
        Remove a split from storage.

        :param split_name: Name of the feature to remove.
        :type split_name: str

        :return: True if the split was found and removed. False otherwise.
        :rtype: bool
        """
        # We need to fetch the split to get the traffic type name prior to deleting.
        fetched = self.get(split_name)
        if fetched is None:
            _LOGGER.warning(
                "Tried to remove feature \"%s\" not present in cache. Ignoring.", split_name
            )
            return

        result = self._uwsgi.cache_del(
            self._KEY_TEMPLATE.format(suffix=split_name),
            _SPLITIO_SPLITS_CACHE_NAMESPACE
        )
        if result is not False:
            _LOGGER.warning("Trying to delete nonexistant split %s. Ignoring.", split_name)

        self._remove_split_from_list(split_name)
        self._decrease_traffic_type_count(fetched.traffic_type_name)

        return result

    def get_change_number(self):
        """
        Retrieve latest split change number.

        :rtype: int
        """
        try:
            return json.loads(self._uwsgi.cache_get(self._KEY_TILL, _SPLITIO_CHANGE_NUMBERS))
        except TypeError:
            return None

    def set_change_number(self, new_change_number):
        """
        Set the latest change number.

        :param new_change_number: New change number.
        :type new_change_number: int
        """
        self._uwsgi.cache_update(self._KEY_TILL, str(new_change_number), 0, _SPLITIO_CHANGE_NUMBERS)

    def get_split_names(self):
        """
        Return a list of all the split names.

        :return: List of split names in cache.
        :rtype: list(str)
        """
        if self._uwsgi.cache_exists(self._KEY_FEATURE_LIST, _SPLITIO_MISC_NAMESPACE):
            try:
                return json.loads(
                    self._uwsgi.cache_get(self._KEY_FEATURE_LIST, _SPLITIO_MISC_NAMESPACE)
                )
            except TypeError:  # Thrown by json.loads when passing none
                pass  # Fall back to default return statement (empty list)
        return []

    def get_all_splits(self):
        """
        Return a list of all splits in cache.

        :return: List of splits.
        :rtype: list(splitio.models.splits.Split)
        """
        return [self.get(split_name) for split_name in self.get_split_names()]

    def is_valid_traffic_type(self, traffic_type_name):
        """
        Return whether the traffic type exists in at least one split in cache.

        :param traffic_type_name: Traffic type to validate.
        :type traffic_type_name: str

        :return: True if the traffic type is valid. False otherwise.
        :rtype: bool
        """
        try:
            tts = json.loads(
                self._uwsgi.cache_get(self._KEY_TRAFFIC_TYPES, _SPLITIO_MISC_NAMESPACE)
            )
            return traffic_type_name in tts
        except TypeError:
            return False

    def _add_split_to_list(self, split_name):
        """
        Add a specific split to the list we keep track of.

        :param split_name: Name of the split to add.
        :type split_name: str
        """
        with UWSGILock(self._uwsgi, self._KEY_FEATURE_LIST_LOCK):
            try:
                current = set(json.loads(
                    self._uwsgi.cache_get(self._KEY_FEATURE_LIST, _SPLITIO_MISC_NAMESPACE)
                ))
            except TypeError:
                current = set()
            current.add(split_name)
            self._uwsgi.cache_update(
                self._KEY_FEATURE_LIST,
                json.dumps(list(current)),
                0,
                _SPLITIO_MISC_NAMESPACE
            )

    def _remove_split_from_list(self, split_name):
        """
        Remove a specific split from the list we keep track of.

        :param split_name: Name of the split to remove.
        :type split_name: str
        """
        with UWSGILock(self._uwsgi, self._KEY_FEATURE_LIST_LOCK):
            try:
                current = set(json.loads(
                    self._uwsgi.cache_get(self._KEY_FEATURE_LIST, _SPLITIO_MISC_NAMESPACE)
                ))
                current.remove(split_name)
                self._uwsgi.cache_update(
                    self._KEY_FEATURE_LIST,
                    json.dumps(list(current)),
                    0,
                    _SPLITIO_MISC_NAMESPACE
                )
            except TypeError:
                # Split list not found, no need to delete anything
                pass
            except KeyError:
                # Split not found in list. nothing to do.
                pass

    def _increase_traffic_type_count(self, traffic_type_name):
        """
        Increase by 1 the count for a specific traffic type.

        :param traffic_type_name: Traffic type name to increase count.
        :type traffic_type_name: str
        """
        with UWSGILock(self._uwsgi, self._KEY_TRAFFIC_TYPES_LOCK):
            try:
                tts = json.loads(
                    self._uwsgi.cache_get(self._KEY_TRAFFIC_TYPES, _SPLITIO_MISC_NAMESPACE)
                )
                tts[traffic_type_name] = tts.get(traffic_type_name, 0) + 1

            except TypeError:
                tts = {traffic_type_name: 1}

            self._uwsgi.cache_update(
                self._KEY_TRAFFIC_TYPES, json.dumps(tts), 0, _SPLITIO_MISC_NAMESPACE
            )

    def _decrease_traffic_type_count(self, traffic_type_name):
        """
        Decreaase by 1 the count for a specific traffic type.

        :param traffic_type_name: Traffic type name to decrease count.
        :type traffic_type_name: str
        """
        with UWSGILock(self._uwsgi, self._KEY_TRAFFIC_TYPES_LOCK):
            try:
                tts = json.loads(
                    self._uwsgi.cache_get(self._KEY_TRAFFIC_TYPES, _SPLITIO_MISC_NAMESPACE)
                )
                tts[traffic_type_name] = tts.get(traffic_type_name, 0) - 1
                if tts[traffic_type_name] <= 0:
                    del tts[traffic_type_name]
            except TypeError:
                # Traffic type list not present. nothing to do here.
                return

            self._uwsgi.cache_update(
                self._KEY_TRAFFIC_TYPES, json.dumps(tts), 0, _SPLITIO_MISC_NAMESPACE
            )

    def kill_locally(self, split_name, default_treatment, change_number):
        """
        Local kill for split

        :param split_name: name of the split to perform kill
        :type split_name: str
        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        if self.get_change_number() > change_number:
            return
        split = self.get(split_name)
        if not split:
            return
        split.local_kill(default_treatment, change_number)
        self.put(split)


class UWSGISegmentStorage(SegmentStorage):
    """UWSGI-Cache based implementation of a split storage."""

    _KEY_TEMPLATE = 'segments.{suffix}'
    _SEGMENT_DATA_KEY_TEMPLATE = 'segmentData.{segment_name}'
    _SEGMENT_CHANGE_NUMBER_KEY_TEMPLATE = 'segment.{segment_name}.till'

    def __init__(self, uwsgi_entrypoint):
        """
        Class constructor.

        :param uwsgi_entrypoint: UWSGI module. Can be the actual module or a mock.
        :type uwsgi_entrypoint: module
        """
        self._uwsgi = uwsgi_entrypoint

    def get(self, segment_name):
        """
        Retrieve a segment.

        :param segment_name: Name of the segment to fetch.
        :type segment_name: str

        :return: Parsed segment if present. None otherwise.
        :rtype: splitio.models.segments.Segment
        """
        key = self._SEGMENT_DATA_KEY_TEMPLATE.format(segment_name=segment_name)
        cn_key = self._SEGMENT_CHANGE_NUMBER_KEY_TEMPLATE.format(segment_name=segment_name)
        try:
            segment_data = json.loads(self._uwsgi.cache_get(key, _SPLITIO_SEGMENTS_CACHE_NAMESPACE))
            change_number = json.loads(self._uwsgi.cache_get(cn_key, _SPLITIO_CHANGE_NUMBERS))
            return segments.from_raw({
                'name': segment_name,
                'added': segment_data,
                'removed': [],
                'till': change_number
            })
        except TypeError:
            _LOGGER.warning(
                "Trying to retrieve nonexistant segment %s. Ignoring.",
                segment_name
            )
            return None

    def update(self, segment_name, to_add, to_remove, change_number=None):
        """
        Update a segment.

        :param segment_name: Name of the segment to update.
        :type segment_name: str
        :param to_add: List of members to add to the segment.
        :type to_add: list
        :param to_remove: List of members to remove from the segment.
        :type to_remove: list
        """
        key = self._SEGMENT_DATA_KEY_TEMPLATE.format(segment_name=segment_name)
        try:
            segment_data = json.loads(self._uwsgi.cache_get(key, _SPLITIO_SEGMENTS_CACHE_NAMESPACE))
        except TypeError:
            segment_data = []
        updated = set(segment_data).union(set(to_add)).difference(to_remove)
        self._uwsgi.cache_update(
            key,
            json.dumps(list(updated)),
            0,
            _SPLITIO_SEGMENTS_CACHE_NAMESPACE
        )

        if change_number is not None:
            self.set_change_number(segment_name, change_number)

    def put(self, segment):
        """
        Put a new segment in storage.

        :param segment: Segment to store.
        :type segment: splitio.models.segments.Segent
        """
        key = self._SEGMENT_DATA_KEY_TEMPLATE.format(segment_name=segment.name)
        self._uwsgi.cache_update(
            key,
            json.dumps(list(segment.keys)),
            0,
            _SPLITIO_SEGMENTS_CACHE_NAMESPACE
        )
        self.set_change_number(segment.name, segment.change_number)

    def get_change_number(self, segment_name):
        """
        Retrieve latest change number for a segment.

        :param segment_name: Name of the segment.
        :type segment_name: str

        :rtype: int
        """
        cnkey = self._SEGMENT_CHANGE_NUMBER_KEY_TEMPLATE.format(segment_name=segment_name)
        try:
            return json.loads(self._uwsgi.cache_get(cnkey, _SPLITIO_CHANGE_NUMBERS))

        except TypeError:
            return None

    def set_change_number(self, segment_name, new_change_number):
        """
        Set the latest change number.

        :param segment_name: Name of the segment.
        :type segment_name: str
        :param new_change_number: New change number.
        :type new_change_number: int
        """
        cn_key = self._SEGMENT_CHANGE_NUMBER_KEY_TEMPLATE.format(segment_name=segment_name)
        self._uwsgi.cache_update(cn_key, json.dumps(new_change_number), 0, _SPLITIO_CHANGE_NUMBERS)

    def segment_contains(self, segment_name, key):
        """
        Check whether a specific key belongs to a segment in storage.

        :param segment_name: Name of the segment to search in.
        :type segment_name: str
        :param key: Key to search for.
        :type key: str

        :return: True if the segment contains the key. False otherwise.
        :rtype: bool
        """
        segment = self.get(segment_name)
        return segment.contains(key)


class UWSGIImpressionStorage(ImpressionStorage):
    """Impressions storage interface."""

    _IMPRESSIONS_KEY = 'SPLITIO.impressions.'
    _LOCK_IMPRESSION_KEY = 'SPLITIO.impressions_lock'
    _IMPRESSIONS_FLUSH = 'SPLITIO.impressions_flush'
    _OVERWRITE_LOCK_SECONDS = 5

    def __init__(self, adapter):
        """
        Class Constructor.

        :param adapter: UWSGI Adapter/Emulator/Module.
        :type: object
        """
        self._uwsgi = adapter

    def put(self, impressions):
        """
        Put one or more impressions in storage.

        :param impressions: List of one or more impressions to store.
        :type impressions: list
        """
        to_store = [i._asdict() for i in impressions]
        with UWSGILock(self._uwsgi, self._LOCK_IMPRESSION_KEY):
            try:
                current = json.loads(self._uwsgi.cache_get(
                    self._IMPRESSIONS_KEY, _SPLITIO_IMPRESSIONS_CACHE_NAMESPACE
                ))
            except TypeError:
                current = []

            self._uwsgi.cache_update(
                self._IMPRESSIONS_KEY,
                json.dumps(current + to_store),
                0,
                _SPLITIO_IMPRESSIONS_CACHE_NAMESPACE
            )

    def pop_many(self, count):
        """
        Pop the oldest N impressions from storage.

        :param count: Number of impressions to pop.
        :type count: int
        """
        with UWSGILock(self._uwsgi, self._LOCK_IMPRESSION_KEY):
            try:
                current = json.loads(self._uwsgi.cache_get(
                    self._IMPRESSIONS_KEY, _SPLITIO_IMPRESSIONS_CACHE_NAMESPACE
                ))
            except TypeError:
                return []

            self._uwsgi.cache_update(
                self._IMPRESSIONS_KEY,
                json.dumps(current[count:]),
                0,
                _SPLITIO_IMPRESSIONS_CACHE_NAMESPACE
            )

        return [
            Impression(
                impression['matching_key'],
                impression['feature_name'],
                impression['treatment'],
                impression['label'],
                impression['change_number'],
                impression['bucketing_key'],
                impression['time']
            ) for impression in current[:count]
        ]

    def request_flush(self):
        """Set a marker in the events cache to indicate that a flush has been requested."""
        self._uwsgi.cache_set(self._IMPRESSIONS_FLUSH, 'ok', 0, _SPLITIO_LOCK_CACHE_NAMESPACE)

    def should_flush(self):
        """
        Return True if a flush has been requested.

        :return: Whether a flush has been requested.
        :rtype: bool
        """
        value = self._uwsgi.cache_get(self._IMPRESSIONS_FLUSH, _SPLITIO_LOCK_CACHE_NAMESPACE)
        return True if value is not None else False

    def acknowledge_flush(self):
        """Acknowledge that a flush has been requested."""
        self._uwsgi.cache_del(self._IMPRESSIONS_FLUSH, _SPLITIO_LOCK_CACHE_NAMESPACE)


class UWSGIEventStorage(EventStorage):
    """Events storage interface."""

    _EVENTS_KEY = 'events'
    _LOCK_EVENTS_KEY = 'events_lock'
    _EVENTS_FLUSH = 'events_flush'
    _OVERWRITE_LOCK_SECONDS = 5

    def __init__(self, adapter):
        """
        Class Constructor.

        :param adapter: UWSGI Adapter/Emulator/Module.
        :type: object
        """
        self._uwsgi = adapter

    def put(self, events):
        """
        Put one or more events in storage.

        :param events: List of one or more events to store.
        :type events: list
        """
        with UWSGILock(self._uwsgi, self._LOCK_EVENTS_KEY):
            try:
                current = json.loads(self._uwsgi.cache_get(
                    self._EVENTS_KEY, _SPLITIO_EVENTS_CACHE_NAMESPACE
                ))
            except TypeError:
                current = []
            self._uwsgi.cache_update(
                self._EVENTS_KEY,
                json.dumps(current + [e.event._asdict() for e in events]),
                0,
                _SPLITIO_EVENTS_CACHE_NAMESPACE
            )

    def pop_many(self, count):
        """
        Pop the oldest N events from storage.

        :param count: Number of events to pop.
        :type count: int
        """
        with UWSGILock(self._uwsgi, self._LOCK_EVENTS_KEY):
            try:
                current = json.loads(self._uwsgi.cache_get(
                    self._EVENTS_KEY, _SPLITIO_EVENTS_CACHE_NAMESPACE
                ))
            except TypeError:
                return []

            self._uwsgi.cache_update(
                self._EVENTS_KEY,
                json.dumps(current[count:]),
                0,
                _SPLITIO_EVENTS_CACHE_NAMESPACE
            )

        return [
            Event(
                event['key'],
                event['traffic_type_name'],
                event['event_type_id'],
                event['value'],
                event['timestamp'],
                event['properties']
            )
            for event in current[:count]
        ]

    def request_flush(self):
        """Set a marker in the events cache to indicate that a flush has been requested."""
        self._uwsgi.cache_set(self._EVENTS_FLUSH, 'requested', 0, _SPLITIO_LOCK_CACHE_NAMESPACE)

    def should_flush(self):
        """
        Return True if a flush has been requested.

        :return: Whether a flush has been requested.
        :rtype: bool
        """
        value = self._uwsgi.cache_get(self._EVENTS_FLUSH, _SPLITIO_LOCK_CACHE_NAMESPACE)
        return True if value is not None else False

    def acknowledge_flush(self):
        """Acknowledge that a flush has been requested."""
        self._uwsgi.cache_del(self._EVENTS_FLUSH, _SPLITIO_LOCK_CACHE_NAMESPACE)


class UWSGITelemetryStorage(TelemetryStorage):
    """Telemetry storage interface."""

    _LATENCIES_KEY = 'SPLITIO.latencies'
    _GAUGES_KEY = 'SPLITIO.gauges'
    _COUNTERS_KEY = 'SPLITIO.counters'

    _LATENCIES_LOCK_KEY = 'SPLITIO.latencies.lock'
    _GAUGES_LOCK_KEY = 'SPLITIO.gauges.lock'
    _COUNTERS_LOCK_KEY = 'SPLITIO.counters.lock'

    def __init__(self, uwsgi_entrypoint):
        """
        Class constructor.

        :param uwsgi_entrypoint: uwsgi module/emulator
        :type uwsgi_entrypoint: object
        """
        self._uwsgi = uwsgi_entrypoint

    def inc_latency(self, name, bucket):
        """
        Add a latency.

        :param name: Name of the latency metric.
        :type name: str
        :param value: Value of the latency metric.
        :tyoe value: int
        """
        if not 0 <= bucket <= 21:
            _LOGGER.error('Incorect bucket "%d" for latency "%s". Ignoring.', bucket, name)
            return

        with UWSGILock(self._uwsgi, self._LATENCIES_LOCK_KEY):
            latencies_raw = self._uwsgi.cache_get(
                self._LATENCIES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            latencies = json.loads(latencies_raw) if latencies_raw else {}
            to_update = latencies.get(name, [0] * 22)
            to_update[bucket] += 1
            latencies[name] = to_update
            self._uwsgi.cache_set(
                self._LATENCIES_KEY,
                json.dumps(latencies),
                0,
                _SPLITIO_METRICS_CACHE_NAMESPACE
            )

    def inc_counter(self, name):
        """
        Increment a counter.

        :param name: Name of the counter metric.
        :type name: str
        """
        with UWSGILock(self._uwsgi, self._COUNTERS_LOCK_KEY):
            counters_raw = self._uwsgi.cache_get(
                self._COUNTERS_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            counters = json.loads(counters_raw) if counters_raw else {}
            value = counters.get(name, 0)
            value += 1
            counters[name] = value
            self._uwsgi.cache_set(
                self._COUNTERS_KEY,
                json.dumps(counters),
                0,
                _SPLITIO_METRICS_CACHE_NAMESPACE
            )

    def put_gauge(self, name, value):
        """
        Add a gauge metric.

        :param name: Name of the gauge metric.
        :type name: str
        :param value: Value of the gauge metric.
        :type value: int
        """
        with UWSGILock(self._uwsgi, self._GAUGES_LOCK_KEY):
            gauges_raw = self._uwsgi.cache_get(self._GAUGES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            gauges = json.loads(gauges_raw) if gauges_raw else {}
            gauges[name] = value
            self._uwsgi.cache_set(
                self._GAUGES_KEY,
                json.dumps(gauges),
                0,
                _SPLITIO_METRICS_CACHE_NAMESPACE
            )

    def pop_counters(self):
        """
        Get all the counters.

        :rtype: list
        """
        with UWSGILock(self._uwsgi, self._COUNTERS_LOCK_KEY):
            counters_raw = self._uwsgi.cache_get(
                self._COUNTERS_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            self._uwsgi.cache_del(self._COUNTERS_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            return json.loads(counters_raw) if counters_raw else {}

    def pop_gauges(self):
        """
        Get all the gauges.

        :rtype: list

        """
        with UWSGILock(self._uwsgi, self._GAUGES_LOCK_KEY):
            gauges_raw = self._uwsgi.cache_get(self._GAUGES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            self._uwsgi.cache_del(self._GAUGES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            return json.loads(gauges_raw) if gauges_raw else {}

    def pop_latencies(self):
        """
        Get all latencies.

        :rtype: list
        """
        with UWSGILock(self._uwsgi, self._LATENCIES_LOCK_KEY):
            latencies_raw = self._uwsgi.cache_get(
                self._LATENCIES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            self._uwsgi.cache_del(self._LATENCIES_KEY, _SPLITIO_METRICS_CACHE_NAMESPACE)
            return json.loads(latencies_raw) if latencies_raw else {}



"""String matchers module."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging
import json
import re
from future.utils import python_2_unicode_compatible
from six import string_types

from splitio.models.grammar.matchers.base import Matcher


_LOGGER = logging.getLogger(__name__)


class Sanitizer(object):  # pylint: disable=too-few-public-methods
    """Numeric input sanitizer."""

    @classmethod
    def ensure_string(cls, data):
        """
        Do a best effort attempt to conver input to a string.

        :param input: user supplied input.
        :type input: mixed.

        :return: String or None
        :rtype: string
        """
        if data is None:  # Failed to fetch attribute. no need to convert.
            return None

        if isinstance(data, string_types):
            return data

        _LOGGER.warning(
            'Supplied attribute is of type %s and should have been a string. ',
            type(data)
        )
        try:
            return json.dumps(data)
        except TypeError:
            return None


class WhitelistMatcher(Matcher):
    """Matcher that returns true if the user key is within a whitelist."""

    def _build(self, raw_matcher):
        """
        Build an WhitelistMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = Sanitizer.ensure_string(self._get_matcher_input(key, attributes))
        if matching_data is None:
            return False
        return matching_data in self._whitelist

    def _add_matcher_specific_properties_to_json(self):
        """Return Whitelist specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'in whitelist [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class StartsWithMatcher(Matcher):
    """Matcher that returns true if the key is a prefix of the stored value."""

    def _build(self, raw_matcher):
        """
        Build an StartsWithMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = Sanitizer.ensure_string(self._get_matcher_input(key, attributes))
        if matching_data is None:
            return False
        return (isinstance(key, string_types) and
                any(matching_data.startswith(s) for s in self._whitelist))

    def _add_matcher_specific_properties_to_json(self):
        """Return StartsWith specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'has one of the following prefixes [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class EndsWithMatcher(Matcher):
    """Matcher that returns true if the key ends with the suffix stored in matcher data."""

    def _build(self, raw_matcher):
        """
        Build an EndsWithMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = Sanitizer.ensure_string(self._get_matcher_input(key, attributes))
        if matching_data is None:
            return False
        return (isinstance(key, string_types) and
                any(matching_data.endswith(s) for s in self._whitelist))

    def _add_matcher_specific_properties_to_json(self):
        """Return EndsWith specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'has one of the following suffixes [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class ContainsStringMatcher(Matcher):
    """Matcher that returns true if the input key is part of the string in matcher data."""

    def _build(self, raw_matcher):
        """
        Build a ContainsStringMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._whitelist = frozenset(raw_matcher['whitelistMatcherData']['whitelist'])

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = Sanitizer.ensure_string(self._get_matcher_input(key, attributes))
        if matching_data is None:
            return False
        return (isinstance(matching_data, string_types) and
                any(s in matching_data for s in self._whitelist))

    def _add_matcher_specific_properties_to_json(self):
        """Return ContainsString specific properties."""
        return {
            'whitelistMatcherData': {
                'whitelist': list(self._whitelist)
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'contains one of the following string: [{whitelist}]'.format(
            whitelist=','.join('"{}"'.format(item) for item in self._whitelist)
        )


class RegexMatcher(Matcher):
    """Matcher that returns true if the user input matches the regex stored in the matcher."""

    def _build(self, raw_matcher):
        """
        Build a RegexMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._data = raw_matcher['stringMatcherData']
        self._regex = re.compile(self._data)

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        matching_data = Sanitizer.ensure_string(self._get_matcher_input(key, attributes))
        if matching_data is None:
            return False
        try:
            matches = re.search(self._regex, matching_data)
            return matches is not None
        except TypeError:
            return False

    def _add_matcher_specific_properties_to_json(self):
        """Return Regex specific properties."""
        return {'stringMatcherData': self._data}



"""Split evaluator module."""
import logging
import threading
from collections import defaultdict, namedtuple
from enum import Enum

import six

from splitio.models.impressions import Impression
from splitio.engine.hashfns import murmur_128
from splitio.engine.cache.lru import SimpleLruCache
from splitio.client.listener import ImpressionListenerException
from splitio import util


_TIME_INTERVAL_MS = 3600 * 1000  # one hour
_IMPRESSION_OBSERVER_CACHE_SIZE = 500000


class ImpressionsMode(Enum):
    """Impressions tracking mode."""

    OPTIMIZED = "OPTIMIZED"
    DEBUG = "DEBUG"


def truncate_time(timestamp_ms):
    """
    Truncate a timestamp in milliseconds to have hour granularity.

    :param timestamp_ms: timestamp generated in the impression.
    :type timestamp_ms: int

    :returns: a timestamp with hour, min, seconds, and ms set to 0.
    :rtype: int
    """
    return timestamp_ms - (timestamp_ms % _TIME_INTERVAL_MS)


class Hasher(object):  # pylint:disable=too-few-public-methods
    """Impression hasher."""

    _PATTERN = "%s:%s:%s:%s:%d"

    def __init__(self, hash_fn=murmur_128, seed=0):
        """
        Class constructor.

        :param hash_fn: Hash function to apply (str, int) -> int
        :type hash_fn: callable

        :param seed: seed to be provided when hashing
        :type seed: int
        """
        self._hash_fn = hash_fn
        self._seed = seed

    def _stringify(self, impression):
        """
        Stringify an impression.

        :param impression: Impression to stringify using _PATTERN
        :type impression: splitio.models.impressions.Impression

        :returns: a string representation of the impression
        :rtype: str
        """
        return self._PATTERN % (impression.matching_key if impression.matching_key else 'UNKNOWN',
                                impression.feature_name if impression.feature_name else 'UNKNOWN',
                                impression.treatment if impression.treatment else 'UNKNOWN',
                                impression.label if impression.label else 'UNKNOWN',
                                impression.change_number if impression.change_number else 0)

    def process(self, impression):
        """
        Hash an impression.

        :param impression: Impression to hash.
        :type impression: splitio.models.impressions.Impression

        :returns: a hash of the supplied impression's relevant fields.
        :rtype: int
        """
        return self._hash_fn(self._stringify(impression), self._seed)


class Observer(object):  # pylint:disable=too-few-public-methods
    """Observe impression and add a previous time if applicable."""

    def __init__(self, size):
        """Class constructor."""
        self._hasher = Hasher()
        self._cache = SimpleLruCache(size)

    def test_and_set(self, impression):
        """
        Examine an impression to determine and set it's previous time accordingly.

        :param impression: Impression to track
        :type impression: splitio.models.impressions.Impression

        :returns: Impression with populated previous time
        :rtype: splitio.models.impressions.Impression
        """
        previous_time = self._cache.test_and_set(self._hasher.process(impression), impression.time)
        return Impression(impression.matching_key,
                          impression.feature_name,
                          impression.treatment,
                          impression.label,
                          impression.change_number,
                          impression.bucketing_key,
                          impression.time,
                          previous_time)


class Counter(object):
    """Class that counts impressions per timeframe."""

    CounterKey = namedtuple('Count', ['feature', 'timeframe'])
    CountPerFeature = namedtuple('CountPerFeature', ['feature', 'timeframe', 'count'])

    def __init__(self):
        """Class constructor."""
        self._data = defaultdict(lambda: 0)
        self._lock = threading.Lock()

    def track(self, impressions, inc=1):
        """
        Register N new impressions for a feature in a specific timeframe.

        :param impressions: generated impressions
        :type impressions: list[splitio.models.impressions.Impression]

        :param inc: amount to increment (defaults to 1)
        :type inc: int
        """
        keys = [Counter.CounterKey(i.feature_name, truncate_time(i.time)) for i in impressions]
        with self._lock:
            for key in keys:
                self._data[key] += inc

    def pop_all(self):
        """
        Clear and return all the counters currently stored.

        :returns: List of count per feature/timeframe objects
        :rtype: list[ImpressionCounter.CountPerFeature]
        """
        with self._lock:
            old = self._data
            self._data = defaultdict(lambda: 0)

        return [Counter.CountPerFeature(k.feature, k.timeframe, v)
                for (k, v) in six.iteritems(old)]


class Manager(object):  # pylint:disable=too-few-public-methods
    """Impression manager."""

    def __init__(self, forwarder, mode=ImpressionsMode.OPTIMIZED, standalone=True, listener=None):
        """
        Construct a manger to track and forward impressions to the queue.

        :param forwarder: function accepting a list of impressions to be added to the queue.
        :type forwarder: callable[list[splitio.models.impressions.Impression]]

        :param mode: Impressions capturing mode.
        :type mode: ImpressionsMode

        :param standalone: whether the SDK is running in standalone sending impressions by itself
        :type standalone: bool

        :param listener: Optional impressions listener that will capture all seen impressions.
        :type listener: splitio.client.listener.ImpressionListenerWrapper
        """
        self._forwarder = forwarder
        self._observer = Observer(_IMPRESSION_OBSERVER_CACHE_SIZE) if standalone else None
        self._counter = Counter() if standalone and mode == ImpressionsMode.OPTIMIZED else None
        self._listener = listener

    def track(self, impressions):
        """
        Track impressions.

        Impressions are analyzed to see if they've been seen before and counted.

        :param impressions: List of impression objects with attributes
        :type impressions: list[tuple[splitio.models.impression.Impression, dict]]
        """
        imps = [(self._observer.test_and_set(imp), attrs) for imp, attrs in impressions] if self._observer \
                else impressions

        if self._counter:
            self._counter.track([imp for imp, _ in imps])

        self._send_impressions_to_listener(imps)

        this_hour = truncate_time(util.utctime_ms())
        self._forwarder([imp for imp, _ in imps] if self._counter is None
                        else [i for i, _ in imps if i.previous_time is None or i.previous_time < this_hour])

    def get_counts(self):
        """
        Return counts of impressions per features.

        :returns: A list of counter objects.
        :rtype: list[Counter.CountPerFeature]
        """
        return self._counter.pop_all() if self._counter is not None else []

    def _send_impressions_to_listener(self, impressions):
        """
        Send impression result to custom listener.

        :param impressions: List of impression objects with attributes
        :type impressions: list[tuple[splitio.models.impression.Impression, dict]]
        """
        if self._listener is not None:
            try:
                for impression, attributes in impressions:
                    self._listener.log_impression(impression, attributes)
            except ImpressionListenerException:
                pass
#                self._logger.error('An exception was raised while calling user-custom impression listener')
#                self._logger.debug('Error', exc_info=True)



"""Events API module."""
import logging

from future.utils import raise_from

from splitio.api import APIException, headers_from_metadata
from splitio.api.client import HttpClientException


_LOGGER = logging.getLogger(__name__)


class EventsAPI(object):  # pylint: disable=too-few-public-methods
    """Class that uses an httpClient to communicate with the events API."""

    def __init__(self, http_client, apikey, sdk_metadata):
        """
        Class constructor.

        :param http_client: HTTP Client responsble for issuing calls to the backend.
        :type http_client: HttpClient
        :param apikey: User apikey token.
        :type apikey: string
        :param sdk_metadata: SDK version & machine name & IP.
        :type sdk_metadata: splitio.client.util.SdkMetadata
        """
        self._client = http_client
        self._apikey = apikey
        self._metadata = headers_from_metadata(sdk_metadata)

    @staticmethod
    def _build_bulk(events):
        """
        Build event bulk as expected by the API.

        :param events: Events to be bundled.
        :type events: list(splitio.models.events.Event)

        :return: Formatted bulk.
        :rtype: dict
        """
        return [
            {
                'key': event.key,
                'trafficTypeName': event.traffic_type_name,
                'eventTypeId': event.event_type_id,
                'value': event.value,
                'timestamp': event.timestamp,
                'properties': event.properties,
            }
            for event in events
        ]

    def flush_events(self, events):
        """
        Send events to the backend.

        :param events: Events bulk
        :type events: list

        :return: True if flush was successful. False otherwise
        :rtype: bool
        """
        bulk = self._build_bulk(events)
        try:
            response = self._client.post(
                'events',
                '/events/bulk',
                self._apikey,
                body=bulk,
                extra_headers=self._metadata
            )
            if not 200 <= response.status_code < 300:
                raise APIException(response.body, response.status_code)
        except HttpClientException as exc:
            _LOGGER.error('Error posting events because an exception was raised by the HTTPClient')
            _LOGGER.debug('Error: ', exc_info=True)
            raise_from(APIException('Events not flushed properly.'), exc)



"""Impressions syncrhonization task."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging

from six.moves import queue

from splitio.tasks import BaseSynchronizationTask
from splitio.tasks.util.asynctask import AsyncTask


_LOGGER = logging.getLogger(__name__)


class ImpressionsSyncTask(BaseSynchronizationTask):
    """Impressions synchronization task uses an asynctask.AsyncTask to send impressions."""

    def __init__(self, synchronize_impressions, period):
        """
        Class constructor.

        :param synchronize_impressions: sender
        :type synchronize_impressions: func
        :param period: How many seconds to wait between subsequent impressions pushes to the BE.
        :type period: int

        """
        self._period = period
        self._task = AsyncTask(synchronize_impressions, self._period,
                               on_stop=synchronize_impressions)

    def start(self):
        """Start executing the impressions synchronization task."""
        self._task.start()

    def stop(self, event=None):
        """Stop executing the impressions synchronization task."""
        self._task.stop(event)

    def is_running(self):
        """
        Return whether the task is running or not.

        :return: True if the task is running. False otherwise.
        :rtype: bool
        """
        return self._task.running()

    def flush(self):
        """Flush impressions in storage."""
        _LOGGER.debug('Forcing flush execution for impressions')
        self._task.force_execution()


class ImpressionsCountSyncTask(BaseSynchronizationTask):
    """Impressions synchronization task uses an asynctask.AsyncTask to send impressions."""

    _PERIOD = 1800  # 30 * 60 # 30 minutes

    def __init__(self, synchronize_counters):
        """
        Class constructor.

        :param synchronize_counters: Handler
        :type synchronize_counters: func

        """
        self._task = AsyncTask(synchronize_counters, self._PERIOD, on_stop=synchronize_counters)

    def start(self):
        """Start executing the impressions synchronization task."""
        self._task.start()

    def stop(self, event=None):
        """Stop executing the impressions synchronization task."""
        self._task.stop(event)

    def is_running(self):
        """
        Return whether the task is running or not.

        :return: True if the task is running. False otherwise.
        :rtype: bool
        """
        return self._task.running()

    def flush(self):
        """Flush impressions in storage."""
        self._task.force_execution()



"""Split Worker tests."""
import time
import queue
import pytest

from splitio.api import APIException
from splitio.push.segmentworker import SegmentWorker
from splitio.models.notification import SegmentChangeNotification

change_number_received = None
segment_name_received = None


def handler_sync(segment_name, change_number):
    global change_number_received
    global segment_name_received
    change_number_received = change_number
    segment_name_received = segment_name
    return


class SegmentWorkerTests(object):
    def test_on_error(self):
        q = queue.Queue()

        def handler_sync(change_number):
            raise APIException('some')

        segment_worker = SegmentWorker(handler_sync, q)
        segment_worker.start()
        assert segment_worker.is_running()

        q.put(SegmentChangeNotification('some', 'SEGMENT_UPDATE', 123456789, 'some'))

        with pytest.raises(Exception):
            segment_worker._handler()

        assert segment_worker.is_running()
        assert segment_worker._worker.is_alive()
        segment_worker.stop()
        time.sleep(1)
        assert not segment_worker.is_running()
        assert not segment_worker._worker.is_alive()

    def test_handler(self):
        q = queue.Queue()
        segment_worker = SegmentWorker(handler_sync, q)
        global change_number_received
        assert not segment_worker.is_running()
        segment_worker.start()
        assert segment_worker.is_running()

        q.put(SegmentChangeNotification('some', 'SEGMENT_UPDATE', 123456789, 'some'))

        time.sleep(0.1)
        assert change_number_received == 123456789
        assert segment_name_received == 'some'

        segment_worker.stop()
        assert not segment_worker.is_running()



"""Keys matchers module."""

from future.utils import python_2_unicode_compatible
from splitio.models.grammar.matchers.base import Matcher

class AllKeysMatcher(Matcher):
    """A matcher that always returns True."""

    def _build(self, raw_matcher):
        """
        Build an AllKeysMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        pass

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        return key is not None

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'in segment all'

    def _add_matcher_specific_properties_to_json(self):
        """Add matcher specific properties to base dict before returning it."""
        return {}


class UserDefinedSegmentMatcher(Matcher):
    """Matcher that returns true when the submitted key belongs to a segment."""

    def _build(self, raw_matcher):
        """
        Build an UserDefinedSegmentMatcher.

        :param raw_matcher: raw matcher as fetched from splitChanges response.
        :type raw_matcher: dict
        """
        self._segment_name = raw_matcher['userDefinedSegmentMatcherData']['segmentName']

    def _match(self, key, attributes=None, context=None):
        """
        Evaluate user input against a matcher and return whether the match is successful.

        :param key: User key.
        :type key: str.
        :param attributes: Custom user attributes.
        :type attributes: dict.
        :param context: Evaluation context
        :type context: dict

        :returns: Wheter the match is successful.
        :rtype: bool
        """
        segment_storage = context.get('segment_storage')
        if not segment_storage:
            raise Exception('Segment storage not present in matcher context.')

        matching_data = self._get_matcher_input(key, attributes)
        if matching_data is None:
            return False
        return segment_storage.segment_contains(self._segment_name, matching_data)

    def _add_matcher_specific_properties_to_json(self):
        """Return UserDefinedSegment specific properties."""
        return {
            'userDefinedSegmentMatcherData': {
                'segmentName': self._segment_name
            }
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string Representation."""
        return 'in segment {segment_name}'.format(
            segment_name=self._segment_name
        )



"""Exponential Backoff duration calculator."""


class Backoff(object):
    """Backoff duration calculator."""

    MAX_ALLOWED_WAIT = 30 * 60  # half an hour

    def __init__(self, base=1):
        """
        Class constructor.

        :param base: basic unit to be multiplied on each iteration (seconds)
        :param base: float
        """
        self._base = base
        self._attempt = 0

    def get(self):
        """
        Return the current time to wait and pre-calculate the next one.

        :returns: time to wait until next retry.
        :rtype: float
        """
        to_return = min(self._base * (2 ** self._attempt), self.MAX_ALLOWED_WAIT)
        self._attempt += 1
        return to_return

    def reset(self):
        """Reset the attempt count."""
        self._attempt = 0



"""Datatypes converters for matchers."""

def ts_truncate_seconds(timestamp):
    """
    Set seconds to zero in a timestamp.

    :param ts: Timestamp in seconds.
    :type ts: int

    :return: Timestamp in seconds, but without counting them (ie: DD-MM-YY HH:MM:00)
    :rtype: int
    """
    return timestamp - (timestamp % 60)

def ts_truncate_time(timestamp):
    """
    Set time to zero in a timestamp.

    :param ts: Timestamp in seconds.
    :type ts: int

    :return: Timestamp in seconds, without counting time (ie: DD-MM-YYYY 00:00:00)
    :rtype: int
    """
    return timestamp - (timestamp % 86400)

def java_ts_to_secs(java_ts):
    """
    Convert java timestamp into unix timestamp.

    :param java_ts: java timestamp in milliseconds.
    :type java_ts: int

    :return: Timestamp in seconds.
    :rtype: int
    """
    return java_ts / 1000

def java_ts_truncate_seconds(java_ts):
    """
    Set seconds to zero in a timestamp.

    :param ts: Timestamp in seconds.
    :type ts: int

    :return: Timestamp in seconds, but without counting them (ie: DD-MM-YY HH:MM:00)
    :rtype: int
    """
    return ts_truncate_seconds(java_ts_to_secs(java_ts))

def java_ts_truncate_time(java_ts):
    """
    Set time to zero in a timestamp.

    :param ts: Timestamp in seconds.
    :type ts: int

    :return: Timestamp in seconds, without counting time (ie: DD-MM-YYYY 00:00:00)
    :rtype: int
    """
    return ts_truncate_time(java_ts_to_secs(java_ts))



"""SSE Status tracker unit tests."""
#pylint:disable=protected-access,no-self-use,line-too-long
from splitio.push.status_tracker import PushStatusTracker, Status
from splitio.push.parser import ControlType, AblyError, OccupancyMessage, ControlMessage


class StatusTrackerTests(object):
    """Parser tests."""

    def test_initial_status_and_reset(self):
        """Test the initial status is ok and reset() works as expected."""
        tracker = PushStatusTracker()
        assert tracker._occupancy_ok()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP
        assert not tracker._shutdown_expected

        tracker._last_control_message = ControlType.STREAMING_PAUSED
        tracker._publishers['control_pri'] = 0
        tracker._publishers['control_sec'] = 1
        tracker._last_status_propagated = Status.PUSH_NONRETRYABLE_ERROR
        tracker.reset()
        assert tracker._occupancy_ok()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP
        assert not tracker._shutdown_expected

    def test_handling_occupancy(self):
        """Test handling occupancy works properly."""
        tracker = PushStatusTracker()
        assert tracker._occupancy_ok()

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_sec', 123, 0)
        assert tracker.handle_occupancy(message) is None

        # old message
        message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 122, 0)
        assert tracker.handle_occupancy(message) is None

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 124, 0)
        assert tracker.handle_occupancy(message) is Status.PUSH_SUBSYSTEM_DOWN

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 125, 1)
        assert tracker.handle_occupancy(message) is Status.PUSH_SUBSYSTEM_UP

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_sec', 125, 2)
        assert tracker.handle_occupancy(message) is None

    def test_handling_control(self):
        """Test handling incoming control messages."""
        tracker = PushStatusTracker()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP

        message = ControlMessage('control_pri', 123, ControlType.STREAMING_ENABLED)
        assert tracker.handle_control_message(message) is None

        # old message
        message = ControlMessage('control_pri', 122, ControlType.STREAMING_PAUSED)
        assert tracker.handle_control_message(message) is None

        message = ControlMessage('control_pri', 124, ControlType.STREAMING_PAUSED)
        assert tracker.handle_control_message(message) is Status.PUSH_SUBSYSTEM_DOWN

        message = ControlMessage('control_pri', 125, ControlType.STREAMING_ENABLED)
        assert tracker.handle_control_message(message) is Status.PUSH_SUBSYSTEM_UP

        message = ControlMessage('control_pri', 126, ControlType.STREAMING_DISABLED)
        assert tracker.handle_control_message(message) is Status.PUSH_NONRETRYABLE_ERROR

        # test that disabling works as well with streaming paused
        tracker = PushStatusTracker()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP

        message = ControlMessage('control_pri', 124, ControlType.STREAMING_PAUSED)
        assert tracker.handle_control_message(message) is Status.PUSH_SUBSYSTEM_DOWN

        message = ControlMessage('control_pri', 126, ControlType.STREAMING_DISABLED)
        assert tracker.handle_control_message(message) is Status.PUSH_NONRETRYABLE_ERROR

    def test_control_occupancy_overlap(self):
        """Test control and occupancy messages together."""
        tracker = PushStatusTracker()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP

        message = ControlMessage('control_pri', 122, ControlType.STREAMING_PAUSED)
        assert tracker.handle_control_message(message) is Status.PUSH_SUBSYSTEM_DOWN

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_sec', 123, 0)
        assert tracker.handle_occupancy(message) is None

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 124, 0)
        assert tracker.handle_occupancy(message) is None

        message = ControlMessage('control_pri', 125, ControlType.STREAMING_ENABLED)
        assert tracker.handle_control_message(message) is None

        message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 126, 1)
        assert tracker.handle_occupancy(message) is Status.PUSH_SUBSYSTEM_UP

    def test_ably_error(self):
        """Test the status tracker reacts appropriately to an ably error."""
        tracker = PushStatusTracker()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP

        message = AblyError(39999, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is None

        message = AblyError(50000, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is None

        tracker.reset()
        message = AblyError(40140, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is Status.PUSH_RETRYABLE_ERROR

        tracker.reset()
        message = AblyError(40149, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is Status.PUSH_RETRYABLE_ERROR

        tracker.reset()
        message = AblyError(40150, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is Status.PUSH_NONRETRYABLE_ERROR

        tracker.reset()
        message = AblyError(40139, 100, 'some message', 'http://somewhere')
        assert tracker.handle_ably_error(message) is Status.PUSH_NONRETRYABLE_ERROR

    def test_disconnect_expected(self):
        """Test that no error is propagated when a disconnect is expected."""
        tracker = PushStatusTracker()
        assert tracker._last_control_message == ControlType.STREAMING_ENABLED
        assert tracker._last_status_propagated == Status.PUSH_SUBSYSTEM_UP
        tracker.notify_sse_shutdown_expected()

        assert tracker.handle_ably_error(AblyError(40139, 100, 'some message', 'http://somewhere')) is None
        assert tracker.handle_ably_error(AblyError(40149, 100, 'some message', 'http://somewhere')) is None
        assert tracker.handle_ably_error(AblyError(39999, 100, 'some message', 'http://somewhere')) is None

        assert tracker.handle_control_message(ControlMessage('control_pri', 123, ControlType.STREAMING_ENABLED)) is None
        assert tracker.handle_control_message(ControlMessage('control_pri', 124, ControlType.STREAMING_PAUSED)) is None
        assert tracker.handle_control_message(ControlMessage('control_pri', 125, ControlType.STREAMING_DISABLED)) is None

        assert tracker.handle_occupancy(OccupancyMessage('[?occupancy=metrics.publishers]control_sec', 123, 0)) is None
        assert tracker.handle_occupancy(OccupancyMessage('[?occupancy=metrics.publishers]control_sec', 124, 1)) is None



"""Split Worker tests."""

import threading
import time
import pytest

from splitio.api.client import HttpResponse
from splitio.api import APIException
from splitio.engine.impressions import Manager as ImpressionsManager
from splitio.engine.impressions import Counter
from splitio.sync.impression import ImpressionsCountSynchronizer
from splitio.api.impressions import ImpressionsAPI


class ImpressionsCountSynchronizerTests(object):
    """ImpressionsCount synchronizer test cases."""

    def test_synchronize_impressions_counts(self, mocker):
        manager = mocker.Mock(spec=ImpressionsManager)

        counters = [
            Counter.CountPerFeature('f1', 123, 2),
            Counter.CountPerFeature('f2', 123, 123),
            Counter.CountPerFeature('f1', 456, 111),
            Counter.CountPerFeature('f2', 456, 222)
        ]

        manager.get_counts.return_value = counters
        api = mocker.Mock(spec=ImpressionsAPI)
        api.flush_counters.return_value = HttpResponse(200, '')
        impression_count_synchronizer = ImpressionsCountSynchronizer(api, manager)
        impression_count_synchronizer.synchronize_counters()

        assert manager.get_counts.mock_calls[0] == mocker.call()
        assert api.flush_counters.mock_calls[0] == mocker.call(counters)

        assert len(api.flush_counters.mock_calls) == 1



"""Redis storage end to end tests."""
#pylint: disable=no-self-use,protected-access,line-too-long,too-few-public-methods

import json
import os

from splitio.client.util import get_metadata
from splitio.models import splits, impressions, events
from splitio.storage.redis import RedisSplitStorage, RedisSegmentStorage, RedisImpressionsStorage, \
    RedisEventsStorage, RedisTelemetryStorage
from splitio.storage.adapters.redis import _build_default_client
from splitio.client.config import DEFAULT_CONFIG


class SplitStorageTests(object):
    """Redis Split storage e2e tests."""

    def test_put_fetch(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            storage = RedisSplitStorage(adapter)
            with open(os.path.join(os.path.dirname(__file__), 'files', 'split_changes.json'), 'r') as flo:
                split_changes = json.load(flo)

            split_objects = [splits.from_raw(raw) for raw in split_changes['splits']]
            for split_object in split_objects:
                raw = split_object.to_json()
                adapter.set(RedisSplitStorage._SPLIT_KEY.format(split_name=split_object.name), json.dumps(raw))
                adapter.incr(RedisSplitStorage._TRAFFIC_TYPE_KEY.format(traffic_type_name=split_object.traffic_type_name))

            original_splits = {split.name: split for split in split_objects}
            fetched_splits = {name: storage.get(name) for name in original_splits.keys()}

            assert set(original_splits.keys()) == set(fetched_splits.keys())

            for original_split in original_splits.values():
                fetched_split = fetched_splits[original_split.name]
                assert original_split.traffic_type_name == fetched_split.traffic_type_name
                assert original_split.seed == fetched_split.seed
                assert original_split.algo == fetched_split.algo
                assert original_split.status == fetched_split.status
                assert original_split.change_number == fetched_split.change_number
                assert original_split.killed == fetched_split.killed
                assert original_split.default_treatment == fetched_split.default_treatment
                for index, original_condition in enumerate(original_split.conditions):
                    fetched_condition = fetched_split.conditions[index]
                    assert original_condition.label == fetched_condition.label
                    assert original_condition.condition_type == fetched_condition.condition_type
                    assert len(original_condition.matchers) == len(fetched_condition.matchers)
                    assert len(original_condition.partitions) == len(fetched_condition.partitions)

            adapter.set(RedisSplitStorage._SPLIT_TILL_KEY, split_changes['till'])
            assert storage.get_change_number() == split_changes['till']

            assert storage.is_valid_traffic_type('user') is True
            assert storage.is_valid_traffic_type('account') is True
            assert storage.is_valid_traffic_type('anything-else') is False

        finally:
            to_delete = [
                "SPLITIO.split.sample_feature",
                "SPLITIO.splits.till",
                "SPLITIO.split.all_feature",
                "SPLITIO.split.killed_feature",
                "SPLITIO.split.Risk_Max_Deductible",
                "SPLITIO.split.whitelist_feature",
                "SPLITIO.split.regex_test",
                "SPLITIO.split.boolean_test",
                "SPLITIO.split.dependency_test",
                "SPLITIO.trafficType.user",
                "SPLITIO.trafficType.account"
            ]
            for item in to_delete:
                adapter.delete(item)

            storage = RedisSplitStorage(adapter)
            assert storage.is_valid_traffic_type('user') is False
            assert storage.is_valid_traffic_type('account') is False

    def test_get_all(self):
        """Test get all names & splits."""
        adapter = _build_default_client({})
        try:
            storage = RedisSplitStorage(adapter)
            with open(os.path.join(os.path.dirname(__file__), 'files', 'split_changes.json'), 'r') as flo:
                split_changes = json.load(flo)

            split_objects = [splits.from_raw(raw) for raw in split_changes['splits']]
            for split_object in split_objects:
                raw = split_object.to_json()
                adapter.set(RedisSplitStorage._SPLIT_KEY.format(split_name=split_object.name), json.dumps(raw))

            original_splits = {split.name: split for split in split_objects}
            fetched_names = storage.get_split_names()
            fetched_splits = {split.name: split for split in storage.get_all_splits()}
            assert set(fetched_names) == set(fetched_splits.keys())

            for original_split in original_splits.values():
                fetched_split = fetched_splits[original_split.name]
                assert original_split.traffic_type_name == fetched_split.traffic_type_name
                assert original_split.seed == fetched_split.seed
                assert original_split.algo == fetched_split.algo
                assert original_split.status == fetched_split.status
                assert original_split.change_number == fetched_split.change_number
                assert original_split.killed == fetched_split.killed
                assert original_split.default_treatment == fetched_split.default_treatment
                for index, original_condition in enumerate(original_split.conditions):
                    fetched_condition = fetched_split.conditions[index]
                    assert original_condition.label == fetched_condition.label
                    assert original_condition.condition_type == fetched_condition.condition_type
                    assert len(original_condition.matchers) == len(fetched_condition.matchers)
                    assert len(original_condition.partitions) == len(fetched_condition.partitions)
        finally:
            adapter.delete(
                'SPLITIO.split.sample_feature',
                'SPLITIO.splits.till',
                'SPLITIO.split.all_feature',
                'SPLITIO.split.killed_feature',
                'SPLITIO.split.Risk_Max_Deductible',
                'SPLITIO.split.whitelist_feature',
                'SPLITIO.split.regex_test',
                'SPLITIO.split.boolean_test',
                'SPLITIO.split.dependency_test'
            )

class SegmentStorageTests(object):
    """Redis Segment storage e2e tests."""

    def test_put_fetch_contains(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            storage = RedisSegmentStorage(adapter)
            adapter.sadd(storage._get_key('some_segment'), 'key1', 'key2', 'key3', 'key4')
            adapter.set(storage._get_till_key('some_segment'), 123)
            assert storage.segment_contains('some_segment', 'key0') is False
            assert storage.segment_contains('some_segment', 'key1') is True
            assert storage.segment_contains('some_segment', 'key2') is True
            assert storage.segment_contains('some_segment', 'key3') is True
            assert storage.segment_contains('some_segment', 'key4') is True
            assert storage.segment_contains('some_segment', 'key5') is False

            fetched = storage.get('some_segment')
            assert fetched.keys == set(['key1', 'key2', 'key3', 'key4'])
            assert fetched.change_number == 123
        finally:
            adapter.delete('SPLITIO.segment.some_segment', 'SPLITIO.segment.some_segment.till')


class ImpressionsStorageTests(object):
    """Redis Impressions storage e2e tests."""

    def _put_impressions(self, adapter, metadata):
        storage = RedisImpressionsStorage(adapter, metadata)
        storage.put([
            impressions.Impression('key1', 'feature1', 'on', 'l1', 123456, 'b1', 321654),
            impressions.Impression('key2', 'feature1', 'on', 'l1', 123456, 'b1', 321654),
            impressions.Impression('key3', 'feature1', 'on', 'l1', 123456, 'b1', 321654)
        ])


    def test_put_fetch_contains(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            self._put_impressions(adapter, get_metadata({}))

            imps = adapter.lrange('SPLITIO.impressions', 0, 2)
            assert len(imps) == 3
            for rawImpression in imps:
                impression = json.loads(rawImpression)
                assert impression['m']['i'] != 'NA'
                assert impression['m']['n'] != 'NA'
        finally:
            adapter.delete('SPLITIO.impressions')

    def test_put_fetch_contains_ip_address_disabled(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            cfg = DEFAULT_CONFIG.copy()
            cfg.update({'IPAddressesEnabled': False})
            self._put_impressions(adapter, get_metadata(cfg))

            imps = adapter.lrange('SPLITIO.impressions', 0, 2)
            assert len(imps) == 3
            for rawImpression in imps:
                impression = json.loads(rawImpression)
                assert impression['m']['i'] == 'NA'
                assert impression['m']['n'] == 'NA'
        finally:
            adapter.delete('SPLITIO.impressions')


class EventsStorageTests(object):
    """Redis Events storage e2e tests."""
    def _put_events(self, adapter, metadata):
        storage = RedisEventsStorage(adapter, metadata)
        storage.put([
            events.EventWrapper(
                event=events.Event('key1', 'user', 'purchase', 3.5, 123456, None),
                size=1024,
            ),
            events.EventWrapper(
                event=events.Event('key2', 'user', 'purchase', 3.5, 123456, None),
                size=1024,
            ),
            events.EventWrapper(
                event=events.Event('key3', 'user', 'purchase', 3.5, 123456, None),
                size=1024,
            ),
        ])

    def test_put_fetch_contains(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            self._put_events(adapter, get_metadata({}))
            evts = adapter.lrange('SPLITIO.events', 0, 2)
            assert len(evts) == 3
            for rawEvent in evts:
                event = json.loads(rawEvent)
                assert event['m']['i'] != 'NA'
                assert event['m']['n'] != 'NA'
        finally:
            adapter.delete('SPLITIO.events')

    def test_put_fetch_contains_ip_address_disabled(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        try:
            cfg = DEFAULT_CONFIG.copy()
            cfg.update({'IPAddressesEnabled': False})
            self._put_events(adapter, get_metadata(cfg))

            evts = adapter.lrange('SPLITIO.events', 0, 2)
            assert len(evts) == 3
            for rawEvent in evts:
                event = json.loads(rawEvent)
                assert event['m']['i'] == 'NA'
                assert event['m']['n'] == 'NA'
        finally:
            adapter.delete('SPLITIO.events')


class TelemetryStorageTests(object):
    """Redis Telemetry storage e2e tests."""

    def test_put_fetch_contains(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        cfg = DEFAULT_CONFIG.copy()
        cfg.update({'IPAddressesEnabled': False})
        metadata = get_metadata(cfg)
        storage = RedisTelemetryStorage(adapter, metadata)
        try:

            storage.inc_counter('counter1')
            storage.inc_counter('counter1')
            storage.inc_counter('counter2')
            assert adapter.get(storage._get_counter_key('counter1')) == '2'
            assert adapter.get(storage._get_counter_key('counter2')) == '1'

            storage.inc_latency('latency1', 3)
            storage.inc_latency('latency1', 3)
            storage.inc_latency('latency2', 6)
            assert adapter.get(storage._get_latency_key('latency1', 3)) == '2'
            assert adapter.get(storage._get_latency_key('latency2', 6)) == '1'

            storage.put_gauge('gauge1', 3)
            storage.put_gauge('gauge2', 1)
            assert adapter.get(storage._get_gauge_key('gauge1')) == '3'
            assert adapter.get(storage._get_gauge_key('gauge2')) == '1'

        finally:
            adapter.delete(
                storage._get_counter_key('counter1'),
                storage._get_counter_key('counter2'),
                storage._get_latency_key('latency1', 3),
                storage._get_latency_key('latency2', 6),
                storage._get_gauge_key('gauge1'),
                storage._get_gauge_key('gauge2')
            )

    def test_put_fetch_contains_ip_address_disabled(self):
        """Test storing and retrieving splits in redis."""
        adapter = _build_default_client({})
        cfg = DEFAULT_CONFIG.copy()
        cfg.update({'IPAddressesEnabled': False})
        metadata = get_metadata(cfg)
        storage = RedisTelemetryStorage(adapter, metadata)
        try:

            storage.inc_counter('counter1')
            storage.inc_counter('counter1')
            storage.inc_counter('counter2')
            assert adapter.get(storage._get_counter_key('counter1')) == '2'
            assert adapter.get(storage._get_counter_key('counter2')) == '1'

            storage.inc_latency('latency1', 3)
            storage.inc_latency('latency1', 3)
            storage.inc_latency('latency2', 6)
            assert adapter.get(storage._get_latency_key('latency1', 3)) == '2'
            assert adapter.get(storage._get_latency_key('latency2', 6)) == '1'

            storage.put_gauge('gauge1', 3)
            storage.put_gauge('gauge2', 1)
            assert adapter.get(storage._get_gauge_key('gauge1')) == '3'
            assert adapter.get(storage._get_gauge_key('gauge2')) == '1'

        finally:
            adapter.delete(
                storage._get_counter_key('counter1'),
                storage._get_counter_key('counter2'),
                storage._get_latency_key('latency1', 3),
                storage._get_latency_key('latency2', 6),
                storage._get_gauge_key('gauge1'),
                storage._get_gauge_key('gauge2')
            )



"""Custom utilities."""

import six


class DynamicDecorator(object):  #pylint: disable=too-few-public-methods
    """
    Decorator that will inject a decorator during class construction.

    This decorator will intercept the __init__(self, *args **kwargs) call,
    and decorate specified methods by instantiating the supplied decorators,
    with arguments extracted and mapped from the constructor call.
    For example:

    def decorator(pos_arg_1, keyword_arg_1=3):
        pass

    @DynamicDecorator(
        decorator,
        ['method1', 'method2'],
        lambda *p, **_: p[0],
        keyword_arg=lambda *_, **kw: kw.get('arg2')
    )
    class SomeClass
        def __init__(self, arg1, arg2=3:
            pass

        def method1(self, x):
            pass

        def method2(self, x):
            pass
    """

    def __init__(self, decorator, methods_to_decorate, *pos_arg_map, **kw_arg_map):
        """
        Construct a decorator with it's mappings.

        :param decorator: Original decorator to apply to specified methods.
        :type decorator: callable
        :param methods_to_decorate: List of methods (strings) where the decorator should be applied
        :type methods_to_decorate: list(string)
        :param pos_arg_map: lambdas to be called with __init__ arguments for the decorator's
            positional arguments.
        :type pos_arg_map: expanded list
        :param kw_arg_map: lambdas to be called with __init__ arguments for the decorator's keyword
            arguments.
        :type kw_arg_map: expanded dict
        """
        self._decorator = decorator
        self._methods = methods_to_decorate
        self._positional_args_lambdas = pos_arg_map
        self._keyword_args_lambdas = kw_arg_map

    def __call__(self, to_decorate):
        """
        Apply the decorator the specified class.

        :param to_decorate: Class to which the decorator will be applied.
        :type to_decorate: class

        :return: a decorated class, which inherits from `to_decorate`
        :rtype: to_decorate
        """
        decorator = self._decorator
        methods = self._methods
        positional_args_lambdas = self._positional_args_lambdas
        keyword_args_lambdas = self._keyword_args_lambdas

        class _decorated(to_decorate):  #pylint: disable=too-few-public-methods
            """
            Decorated class wrapper.

            This wrapper uses the __init__ to catch required arguments,
            instantiate the decorator with the appropriate parameters and then create a child
            class with decorated behaviour.
            """

            def __init__(self, *args, **kwargs):
                """Decorate class constructor."""
                # calculate positional and keyword arguments needed to build the decorator.
                positional = [pos_func(*args, **kwargs) for pos_func in positional_args_lambdas]
                keyword = {
                    key: func(*args, **kwargs)
                    for (key, func) in six.iteritems(keyword_args_lambdas)
                }

                # call original class constructor
                to_decorate.__init__(self, *args, **kwargs)

                # decorate specified methods
                for method in methods:
                    decorated_method = decorator(*positional, **keyword)(getattr(self, method))
                    setattr(to_decorate, method, decorated_method)

        return _decorated



"""
Hash functions module.

This module contains hash functions implemented in pure python
as well as the optional import (if installed) of a C compiled murmur hash
function with python bindings.
"""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from splitio.models.splits import HashAlgorithm
from splitio.engine.hashfns import legacy

try:
    # First attempt to import module with C++ core (faster)
    import mmh3cffi

    def _murmur_hash(key, seed):
        return mmh3cffi.hash_str(key, seed)

    def _murmur_hash128(key, seed):
        return mmh3cffi.hash_str_128(key, seed)[0]

except ImportError:
    # Fallback to interpreted python hash algoritm (slower)
    from splitio.engine.hashfns import murmur3py  #pylint: disable=ungrouped-imports
    _murmur_hash = murmur3py.murmur32_py  #pylint: disable=invalid-name
    _murmur_hash128 = lambda k, s: murmur3py.hash128_x64(k, s)[0] #pylint: disable=invalid-name


_HASH_ALGORITHMS = {
    HashAlgorithm.LEGACY: legacy.legacy_hash,
    HashAlgorithm.MURMUR: _murmur_hash
}

murmur_128 = _murmur_hash128  #pylint: disable=invalid-name

def get_hash_fn(algo):
    """
    Return appropriate hash function for requested algorithm.

    :param algo: Algoritm to use
    :type algo: int
    :return: Hash function
    :rtype: function
    """
    return _HASH_ALGORITHMS.get(algo, legacy.legacy_hash)



"""Split API module."""


class APIException(Exception):
    """Exception to raise when an API call fails."""

    def __init__(self, custom_message, status_code=None):
        """Constructor."""
        Exception.__init__(self, custom_message)
        self._status_code = status_code if status_code else -1

    @property
    def status_code(self):
        """Return HTTP status code."""
        return self._status_code


def headers_from_metadata(sdk_metadata):
    """
    Generate a dict with headers required by data-recording API endpoints.

    :param sdk_metadata: SDK Metadata object, generated at sdk initialization time.
    :type sdk_metadata: splitio.client.util.SdkMetadata

    :return: A dictionary with headers.
    :rtype: dict
    """
    return {
        'SplitSDKVersion': sdk_metadata.sdk_version,
        'SplitSDKMachineIP': sdk_metadata.instance_ip,
        'SplitSDKMachineName': sdk_metadata.instance_name
    } if sdk_metadata.instance_ip != 'NA' and sdk_metadata.instance_ip != 'unknown' else {
        'SplitSDKVersion': sdk_metadata.sdk_version,
    }



"""Localhost client mocked components."""

import itertools
import logging
import re

from future.utils import raise_from
import yaml

from splitio.models import splits
from splitio.storage import ImpressionStorage, EventStorage, TelemetryStorage
from splitio.tasks import BaseSynchronizationTask
from splitio.tasks.util import asynctask

_LEGACY_COMMENT_LINE_RE = re.compile(r'^#.*$')
_LEGACY_DEFINITION_LINE_RE = re.compile(r'^(?<![^#])(?P<feature>[\w_-]+)\s+(?P<treatment>[\w_-]+)$')


_LOGGER = logging.getLogger(__name__)


class LocalhostImpressionsStorage(ImpressionStorage):
    """Impression storage that doesn't cache anything."""

    def put(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def pop_many(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass


class LocalhostEventsStorage(EventStorage):
    """Impression storage that doesn't cache anything."""

    def put(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def pop_many(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass


class LocalhostTelemetryStorage(TelemetryStorage):
    """Impression storage that doesn't cache anything."""

    def inc_latency(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def inc_counter(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def put_gauge(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def pop_latencies(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def pop_counters(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass

    def pop_gauges(self, *_, **__):  # pylint: disable=arguments-differ
        """Accept any arguments and do nothing."""
        pass



"""Asynchronous tasks that can be controlled."""
import threading
import logging

from six.moves import queue


__TASK_STOP__ = 0
__TASK_FORCE_RUN__ = 1

_LOGGER = logging.getLogger(__name__)


def _safe_run(func):
    """
    Execute a function wrapped in a try-except block.

    If anything goes wrong returns false instead of propagating the exception.

    :param func: Function to be executed, receives no arguments and it's return
        value is ignored.
    """
    try:
        func()
        return True
    except Exception:  # pylint: disable=broad-except
        # Catch any exception that might happen to avoid the periodic task
        # from ending and allowing for a recovery, as well as preventing
        # an exception from propagating and breaking the main thread
        _LOGGER.error('Something went wrong when running passed function.')
        _LOGGER.debug('Original traceback:', exc_info=True)
        return False


class AsyncTask(object):  # pylint: disable=too-many-instance-attributes
    """
    Asyncrhonous controllable task class.

    This class creates is used to wrap around a function to treat it as a
    periodic task. This task can be stopped, it's execution can be forced, and
    it's status (whether it's running or not) can be obtained from the task
    object.
    It also allows for "on init" and "on stop" functions to be passed.
    """

    def __init__(self, main, period, on_init=None, on_stop=None):
        """
        Class constructor.

        :param main: Main function to be executed periodically
        :type main: callable
        :param period: How many seconds to wait between executions
        :type period: int
        :param on_init: Function to be executed ONCE before the main one
        :type on_init: callable
        :param on_stop: Function to be executed ONCE after the task has finished
        :type on_stop: callable
        """
        self._on_init = on_init
        self._main = main
        self._on_stop = on_stop
        self._period = period
        self._messages = queue.Queue()
        self._running = False
        self._thread = None
        self._stop_event = None

    def _execution_wrapper(self):
        """
        Execute user defined function in separate thread.

        It will execute the "on init" hook is available. If an exception is
        raised it will abort execution, otherwise it will enter an infinite
        loop in which the main function is executed every <period> seconds.
        After stop has been called the "on stop" hook will be invoked if
        available.

        All custom functions are run within a _safe_run() function which
        prevents exceptions from being propagated.
        """
        try:
            if self._on_init is not None:
                if not _safe_run(self._on_init):
                    _LOGGER.error("Error running task initialization function, aborting execution")
                    self._running = False
                    return
            self._running = True
            while True:
                try:
                    msg = self._messages.get(True, self._period)
                    if msg == __TASK_STOP__:
                        _LOGGER.debug("Stop signal received. finishing task execution")
                        break
                    elif msg == __TASK_FORCE_RUN__:
                        _LOGGER.debug("Force execution signal received. Running now")
                        if not _safe_run(self._main):
                            _LOGGER.error("An error occurred when executing the task. "
                                          "Retrying after perio expires")
                        continue
                except queue.Empty:
                    # If no message was received, the timeout has expired
                    # and we're ready for a new execution
                    pass

                if not _safe_run(self._main):
                    _LOGGER.error(
                        "An error occurred when executing the task. "
                        "Retrying after perio expires"
                    )
        finally:
            self._cleanup()

    def _cleanup(self):
        """Execute on_stop callback, set event if needed, update status."""
        if self._on_stop is not None:
            if not _safe_run(self._on_stop):
                _LOGGER.error("An error occurred when executing the task's OnStop hook. ")

        self._running = False

        if self._stop_event is not None:
            self._stop_event.set()

    def start(self):
        """Start the async task."""
        if self._running:
            _LOGGER.warning("Task is already running. Ignoring .start() call")
            return

        # Start execution
        self._thread = threading.Thread(target=self._execution_wrapper,
                                        name='AsyncTask::' + getattr(self._main, '__name__', 'N/S'))
        self._thread.setDaemon(True)
        try:
            self._thread.start()

        except RuntimeError:
            _LOGGER.error("Couldn't create new thread for async task")
            _LOGGER.debug('Error: ', exc_info=True)

    def stop(self, event=None):
        """
        Send a signal to the thread in order to stop it. If the task is not running do nothing.

        Optionally accept an event to be set upon task completion.

        :param event: Event to set when the task completes.
        :type event: threading.Event
        """
        if event is not None:
            self._stop_event = event

        if not self._running:
            if self._stop_event is not None:
                event.set()
            return

        # Queue is of infinite size, should not raise an exception
        self._messages.put(__TASK_STOP__, False)

    def force_execution(self):
        """Force an execution of the task without waiting for the period to end."""
        if not self._running:
            return
        # Queue is of infinite size, should not raise an exception
        self._messages.put(__TASK_FORCE_RUN__, False)

    def running(self):
        """Return whether the task is running or not."""
        return self._running



"""SSE mock server."""
import json
from collections import namedtuple
import queue
import threading

from http.server import HTTPServer, BaseHTTPRequestHandler


Request = namedtuple('Request', ['method', 'path', 'headers', 'body'])


class SSEMockServer(object):
    """SSE server for testing purposes."""

    protocol_version = 'HTTP/1.1'

    GRACEFUL_REQUEST_END = 'REQ-END'
    VIOLENT_REQUEST_END = 'REQ-KILL'

    def __init__(self, req_queue=None):
        """Consruct a mock server."""
        self._queue = queue.Queue()
        self._server = HTTPServer(('localhost', 0),
                                  lambda *xs: SSEHandler(self._queue, *xs, req_queue=req_queue))
        self._server_thread = threading.Thread(target=self._blocking_run)
        self._server_thread.setDaemon(True)
        self._done_event = threading.Event()

    def _blocking_run(self):
        """Execute."""
        self._server.serve_forever()
        self._done_event.set()

    def port(self):
        """Return the assigned port."""
        return self._server.server_port

    def publish(self, event):
        """Publish an event."""
        self._queue.put(event, block=False)

    def start(self):
        """Start the server asyncrhonously."""
        self._server_thread.start()

    def wait(self, timeout=None):
        """Wait for the server to shutdown."""
        return self._done_event.wait(timeout)

    def stop(self):
        """Stop the server."""
        self._server.shutdown()


class SSEHandler(BaseHTTPRequestHandler):
    """Handler."""

    def __init__(self, event_queue, *args, **kwargs):
        """Construct a handler."""
        self._queue = event_queue
        self._req_queue = kwargs.get('req_queue')
        BaseHTTPRequestHandler.__init__(self, *args)

    def do_GET(self):  #pylint:disable=invalid-name
        """Respond to a GET request."""
        self.send_response(200)
        self.send_header("Content-type", "text/event-stream")
        self.send_header("Transfer-Encoding", "chunked")
        self.send_header("Connection", "keep-alive")
        self.end_headers()

        if self._req_queue is not None:
            headers = dict(zip(self.headers.keys(), self.headers.values()))
            self._req_queue.put(Request('GET', self.path, headers, None))

        def write_chunk(chunk):
            """Write an event/chunk."""
            tosend = '%X\r\n%s\r\n'%(len(chunk), chunk)
            self.wfile.write(tosend.encode('utf-8'))

        while True:
            event = self._queue.get()
            if event == SSEMockServer.GRACEFUL_REQUEST_END:
                break
            elif event == SSEMockServer.VIOLENT_REQUEST_END:
                raise Exception('exploding')

            chunk = ''
            chunk += 'id: % s\n' % event['id'] if 'id' in event else ''
            chunk += 'event: % s\n' % event['event'] if 'event' in event else ''
            chunk += 'retry: % s\n' % event['retry'] if 'retry' in event else ''
            chunk += 'data: % s\n' % event['data'] if 'data' in event else ''
            if chunk != '':
                write_chunk(chunk + '\r\n')

        self.wfile.write('0\r\n\r\n'.encode('utf-8'))


class SplitMockServer(object):
    """SDK server mock for testing purposes."""

    protocol_version = 'HTTP/1.1'

    def __init__(self, split_changes=None, segment_changes=None, req_queue=None,
                 auth_response=None):
        """
        Consruct a mock server.

        :param changes: mapping of changeNumbers to splitChanges responses
        :type changes: dict
        """
        split_changes = split_changes if split_changes is not None else {}
        segment_changes = segment_changes if segment_changes is not None else {}
        self._server = HTTPServer(('localhost', 0),
                                  lambda *xs: SDKHandler(split_changes, segment_changes, *xs,
                                                         req_queue=req_queue,
                                                         auth_response=auth_response))
        self._server_thread = threading.Thread(target=self._blocking_run, name="SplitMockServer")
        self._server_thread.setDaemon(True)
        self._done_event = threading.Event()

    def _blocking_run(self):
        """Execute."""
        self._server.serve_forever()
        self._done_event.set()

    def port(self):
        """Return the assigned port."""
        return self._server.server_port

    def start(self):
        """Start the server asyncrhonously."""
        self._server_thread.start()

    def wait(self, timeout=None):
        """Wait for the server to shutdown."""
        return self._done_event.wait(timeout)

    def stop(self):
        """Stop the server."""
        self._server.shutdown()


class SDKHandler(BaseHTTPRequestHandler):
    """Handler."""

    def __init__(self, split_changes, segment_changes, *args, **kwargs):
        """Construct a handler."""
        self._req_queue = kwargs.get('req_queue')
        self._auth_response = kwargs.get('auth_response')
        self._split_changes = split_changes
        self._segment_changes = segment_changes
        BaseHTTPRequestHandler.__init__(self, *args)

    def _parse_qs(self):
        raw_query = self.path.split('?')[1] if '?' in self.path else ''
        return dict([item.split('=') for item in raw_query.split('&')])

    def _handle_segment_changes(self):
        qstring = self._parse_qs()
        since = int(qstring.get('since', -1))
        name = self.path.split('/')[-1].split('?')[0]
        if name is None:
            self.send_response(400)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write('{}'.encode('utf-8'))
            return

        to_send = self._segment_changes.get((name, since,))
        if to_send is None:
            self.send_response(404)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write('{}'.encode('utf-8'))
            return

        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(to_send).encode('utf-8'))

    def _handle_split_changes(self):
        qstring = self._parse_qs()
        since = int(qstring.get('since', -1))
        to_send = self._split_changes.get(since)
        if to_send is None:
            self.send_response(404)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write('{}'.encode('utf-8'))
            return

        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(to_send).encode('utf-8'))

    def _handle_auth(self):
        if not self._auth_response:
            self.send_response(401)
            self.send_header("Content-type", "application/json")
            self.end_headers()
            self.wfile.write('{}'.encode('utf-8'))
            return

        self.send_response(200)
        self.send_header("Content-type", "application/json")
        self.end_headers()
        self.wfile.write(json.dumps(self._auth_response).encode('utf-8'))

    def do_GET(self):  #pylint:disable=invalid-name
        """Respond to a GET request."""
        if self._req_queue is not None:
            headers = self._format_headers()
            self._req_queue.put(Request('GET', self.path, headers, None))

        if self.path.startswith('/api/splitChanges'):
            self._handle_split_changes()
        elif self.path.startswith('/api/segmentChanges'):
            self._handle_segment_changes()
        elif self.path.startswith('/api/auth'):
            self._handle_auth()
        else:
            self.send_response(404)
            self.send_header("Content-type", "application/json")
            self.end_headers()

    def do_POST(self):  #pylint:disable=invalid-name
        """Respond to a GET request."""
        if self._req_queue is not None:
            headers = self._format_headers()
            length = int(headers.get('content-length'))
            body = self.rfile.read(length) if length else None
            self._req_queue.put(Request('POST', self.path, headers, body))

        if self.path in set(['/api/testImpressions/bulk', '/testImpressions/count',
                             '/api/events/bulk', '/metrics/times', '/metrics/count',
                             '/metrics/gauge']):

            self.send_response(200)
            self.send_header("Content-type", "application/json")
            self.end_headers()
        else:
            self.send_response(404)
            self.send_header("Content-type", "application/json")
            self.end_headers()

    def _format_headers(self):
        """Format headers and return them as a dict."""
        return dict(zip([k.lower() for k in self.headers.keys()], self.headers.values()))
